# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import inspect
import json
import os
import pathlib
import re
import ssl
import tempfile
import typing as t
import warnings
from importlib import util
from types import FrameType

import __main__
import markdown as md_lib
import tzlocal
from flask import Blueprint, Flask, g, jsonify, request, send_from_directory
from werkzeug.utils import secure_filename

if util.find_spec("pyngrok"):
    from pyngrok import ngrok

from ._default_config import default_config
from ._page import _Page
from .config import Config, ConfigParameter, _Config
from .data.content_accessor import _ContentAccessor
from .data.data_accessor import _DataAccessor, _DataAccessors
from .data.data_format import _DataFormat
from .data.data_scope import _DataScopes
from .page import Page
from .partial import Partial
from .renderers import _EmptyPage
from .renderers._markdown import _TaipyMarkdownExtension
from .renderers.jsonencoder import _TaipyJsonEncoder
from .server import _Server
from .state import State
from .types import _WsType
from .utils import (
    _delscopeattr,
    _filter_locals,
    _get_client_var_name,
    _get_module_name_from_frame,
    _get_non_existent_file_path,
    _getscopeattr,
    _getscopeattr_drill,
    _hasscopeattr,
    _is_in_notebook,
    _LocalsContext,
    _MapDict,
    _setscopeattr,
    _setscopeattr_drill,
    _TaipyBase,
    _TaipyContent,
    _TaipyContentImage,
    _TaipyData,
    _TaipyLov,
    _TaipyLovValue,
    _variable_decode,
    _variable_encode,
)
from .utils._adapter import _Adapter
from .utils._bindings import _Bindings
from .utils._evaluator import _Evaluator
from .utils._variable_directory import _RE_TPMDL_DECODE, _VariableDirectory


class Gui:
    """Entry point for the Graphical User Interface generation.

    Attributes:

        on_action (Callable): The function that is called when a control
            triggers an action, as the result of an interaction with the end-user.<br/>
            It defaults to the `on_action()` global function defined in the Python
            application. If there is no such function, actions will not trigger anything.
        on_change (Callable): The function that is called when a control
            modifies variables it is bound to, as the result of an interaction with the
            end-user.<br/>
            It defaults to the `on_change()` global function defined in the Python
            application. If there is no such function, user interactions will not trigger
            anything.
        on_init (Callable): The function that is called on the first connection of a new user.<br/>
            It defaults to the `on_init()` global function defined in the Python
            application.
        state (State^): **Only defined when running in an IPython notebook context.**<br/>
            The unique instance of `State^` that you can use to change bound variables
            directly, potentially impacting the interface in real-time.

    !!! note
        This class belong to and is documented in the `taipy.gui` package but it is
        accessible from the top `taipy` package to simplify its access, allowing to
        use:
        ```py
        from taipy import Gui
        ```
    """

    __root_page_name = "TaiPy_root_page"
    __env_filename = "taipy.gui.env"
    __UI_BLOCK_NAME = "TaipyUiBlockVar"
    __MESSAGE_GROUPING_NAME = "TaipyMessageGrouping"
    __ON_INIT_NAME = "TaipyOnInit"
    __CONTENT_ROOT = "/taipy-content/"
    __UPLOAD_URL = "/taipy-uploads"

    __RE_HTML = re.compile(r"(.*?)\.html")
    __RE_MD = re.compile(r"(.*?)\.md")
    __RE_PAGE_NAME = re.compile(r"^[\w\-\/]+$")

    __reserved_routes: t.List[str] = ["taipy-init", "taipy-jsx", "taipy-content", "taipy-uploads"]
    _aggregate_functions: t.List[str] = ["count", "sum", "mean", "median", "min", "max", "std", "first", "last"]

    __LOCAL_TZ = str(tzlocal.get_localzone())

    def __init__(
        self,
        page: t.Optional[t.Union[str, Page]] = None,
        pages: t.Optional[dict] = None,
        css_file: t.Optional[str] = None,
        path_mapping: t.Optional[dict] = {},
        env_filename: t.Optional[str] = None,
        flask: t.Optional[Flask] = None,
    ):
        """Initialize a new Gui instance.

        Arguments:
            page (Optional[Union[str, Page^]]): An optional `Page^` instance that is used
                when there is a single page in this interface, referenced as the _root_
                page (located at `/`).<br/>
                If _page_ is a raw string and if it holds a path to a readable file then
                a `Markdown^` page is built from the content of that file.<br/>
                If _page_ is a string that does not indicate a path to readable file then
                a `Markdown^` page is built from that string.<br/>
                Note that if _pages_ is provided, those pages are added as well.
            pages (Optional[dict]): Used if you want to initialize this instance with a set
                of pages.<br/>
                The method `(Gui.)add_pages(pages)^` is called if _pages_ is not None.
                You can find details on the possible values of this argument in the
                documentation for this method.
            css_file (str): An optional pathname to a CSS file that gets used as a style sheet in
                all the pages.<br/>
                The default value is a file that has the same base name as the Python
                file defining the `main` function, sitting next to this Python file,
                with the `.css` extension.
            path_mapping (Optional[dict]): A dictionary that associates a URL prefix to
                a path in the server filesystem.</br>
                If the assets of your application are located in _/home/me/app_assets_ and
                you want to access them with just _assets_ in your application, you can
                set _path_mapping={"assets": "/home/me/app_assets"}_. If your application
                then requests the file _"/assets/images/logo.png"_, the server searches
                for the file  _"/home/me/app_assets/images/logo.png"_.
            env_filename (Optional[str]): An optional file from which to load application
                configuration variables (see the
                [Configuration](../gui/configuration.md#configuring-the-gui-instance) section
                of the User Manual for details.)</br>
                The default value is "taipy.gui.env"
            flask (Optional[Flask]): An optional instance of a Flask application object.<br/>
                If this argument is set, this `Gui` instance will use the value of this argument
                as the underlying server. If omitted or set to None, this `Gui` will create its
                own Flask application instance and use it to serve the pages.
        """
        # store suspected local containing frame
        self.__frame = t.cast(FrameType, t.cast(FrameType, inspect.currentframe()).f_back)
        self.__default_module_name = _get_module_name_from_frame(self.__frame)

        # Preserve server config for server initialization
        self._path_mapping = path_mapping
        self._flask = flask
        if css_file is None:
            css_file = os.path.splitext(os.path.basename(self.__frame.f_code.co_filename))[0] or "Taipy"
        self._css_file = css_file

        self._config = _Config()
        self.__content_accessor = None
        self._accessors = _DataAccessors()
        self.__state: t.Optional[State] = None
        self.__bindings = _Bindings(self)
        self.__locals_context = _LocalsContext()
        self.__var_dir = _VariableDirectory(self.__locals_context)

        self.__evaluator: _Evaluator = None  # type: ignore
        self.__adapter = _Adapter()
        self.__directory_name_of_pages: t.List[str] = []

        # default actions
        self.on_action: t.Optional[t.Callable] = None
        self.on_change: t.Optional[t.Callable] = None
        self.on_init: t.Optional[t.Callable] = None

        # sid from client_id
        self.__client_id_2_sid: t.Dict[str, t.Set[str]] = {}

        # Load default config
        self._flask_blueprint: t.List[Blueprint] = []
        self._config._load(default_config)

        # Load Markdown extension
        # NOTE: Make sure, if you change this extension list, that the User Manual gets updated.
        # There's a section that explicitly lists these extensions in
        #      docs/gui/pages.md#markdown-specifics
        self._markdown = md_lib.Markdown(
            extensions=[
                "fenced_code",
                "meta",
                "admonition",
                "sane_lists",
                "tables",
                "attr_list",
                "md_in_html",
                _TaipyMarkdownExtension(gui=self),
            ]
        )

        if page:
            self.add_page(name=Gui.__root_page_name, page=page)
        if pages is not None:
            self.add_pages(pages)
        if env_filename is not None:
            self.__env_filename = env_filename

    def __get_content_accessor(self):
        if self.__content_accessor is None:
            self.__content_accessor = _ContentAccessor(self._get_config("data_url_max_size", 50 * 1024))
        return self.__content_accessor

    def _bindings(self):
        return self.__bindings

    def _get_data_scope(self):
        return self.__bindings._get_data_scope()

    def _get_config(self, name: ConfigParameter, default_value: t.Any) -> t.Any:
        return self._config._get_config(name, default_value)

    def _get_themes(self) -> t.Optional[t.Dict[str, t.Any]]:
        theme = self._get_config("theme", None)
        dark_theme = self._get_config("dark_theme", None)
        light_theme = self._get_config("light_theme", None)
        res = {}
        if theme:
            res["base"] = theme
        if dark_theme:
            res["dark"] = dark_theme
        if light_theme:
            res["light"] = light_theme
        if theme or dark_theme or light_theme:
            return res
        return None

    def _bind(self, name: str, value: t.Any) -> None:
        self._bindings()._bind(name, value)

    def __get_state(self):
        return self.__state

    def _get_client_id(self) -> str:
        return (
            _DataScopes._GLOBAL_ID if self._bindings()._get_single_client() else getattr(g, "client_id", "unknown id")
        )

    def __set_client_id_in_context(self, client_id: t.Optional[str] = None):
        if not client_id and request:
            client_id = request.args.get("client_id", "")
        if client_id and request:
            if sid := getattr(request, "sid", None):
                sids = self.__client_id_2_sid.get(client_id, None)
                if sids is None:
                    sids = set()
                    self.__client_id_2_sid[client_id] = sids
                sids.add(sid)
        g.client_id = client_id

    def __is_var_modified_in_context(self, var_name: str, derived_vars: t.Set[str]) -> bool:
        modified_vars: t.Optional[t.Set[str]] = getattr(g, "modified_vars", None)
        der_vars: t.Optional[t.Set[str]] = getattr(g, "derived_vars", None)
        setattr(g, "update_count", getattr(g, "update_count", 0) + 1)
        if modified_vars is None:
            modified_vars = set()
            g.modified_vars = modified_vars
        if der_vars is None:
            g.derived_vars = derived_vars
        else:
            der_vars.update(derived_vars)
        if var_name in modified_vars:
            return True
        else:
            modified_vars.add(var_name)
            return False

    def __clean_vars_on_exit(self) -> t.Optional[t.Set[str]]:
        update_count = getattr(g, "update_count", 0) - 1
        if update_count < 1:
            vars: t.Set[str] = getattr(g, "derived_vars", set())
            delattr(g, "update_count")
            delattr(g, "modified_vars")
            delattr(g, "derived_vars")
            return vars
        else:
            setattr(g, "update_count", update_count)
            return None

    def _manage_message(self, msg_type: _WsType, message: dict) -> None:
        try:
            self.__set_client_id_in_context(message.get("client_id"))
            self._set_locals_context(message.get("module_context") or None)
            if msg_type == _WsType.UPDATE.value:
                payload = message.get("payload", {})
                self.__front_end_update(
                    str(message.get("name")),
                    payload.get("value"),
                    message.get("propagate", True),
                    payload.get("relvar"),
                    payload.get("on_change"),
                )
            elif msg_type == _WsType.ACTION.value:
                self.__on_action(message.get("name"), message.get("payload"))
            elif msg_type == _WsType.DATA_UPDATE.value:
                self.__request_data_update(str(message.get("name")), message.get("payload"))
            elif msg_type == _WsType.REQUEST_UPDATE.value:
                self.__request_var_update(message.get("payload"))
            elif msg_type == _WsType.CLIENT_ID.value:
                self._bindings()._get_or_create_scope(message.get("payload", ""))
            self._reset_locals_context()
        except Exception as e:
            warnings.warn(f"Decoding Message has failed: {message}\n{e}")

    def __front_end_update(
        self,
        var_name: str,
        value: t.Any,
        propagate=True,
        rel_var: t.Optional[str] = None,
        on_change: t.Optional[str] = None,
    ) -> None:
        if not var_name:
            return
        # Check if Variable is a managed type
        current_value = _getscopeattr_drill(self, self.__evaluator.get_hash_from_expr(var_name))
        if isinstance(current_value, _TaipyData):
            return
        elif rel_var and isinstance(current_value, _TaipyLovValue):  # pragma: no cover
            lov_holder = _getscopeattr_drill(self, self.__evaluator.get_hash_from_expr(rel_var))
            if isinstance(lov_holder, _TaipyLov):
                val = value if isinstance(value, list) else [value]
                elt_4_ids = self.__adapter._get_elt_per_ids(lov_holder.get_name(), lov_holder.get())
                ret_val = [elt_4_ids.get(x, x) for x in val]
                if isinstance(value, list):
                    value = ret_val
                elif ret_val:
                    value = ret_val[0]
        elif isinstance(current_value, _TaipyBase):
            value = current_value.cast_value(value)
        self._update_var(
            var_name, value, propagate, current_value if isinstance(current_value, _TaipyBase) else None, on_change
        )

    def _update_var(
        self,
        var_name: str,
        value: t.Any,
        propagate=True,
        holder: t.Optional[_TaipyBase] = None,
        on_change: t.Optional[str] = None,
        from_map_dict: bool = False,
    ) -> None:
        if from_map_dict:
            var_name = _variable_encode(var_name, self._get_locals_context())
        if holder:
            var_name = holder.get_name()
        hash_expr = self.__evaluator.get_hash_from_expr(var_name)
        derived_vars = {hash_expr}
        # Use custom attrsetter function to allow value binding for _MapDict
        if propagate:
            _setscopeattr_drill(self, hash_expr, value)
            # In case expression == hash (which is when there is only a single variable in expression)
            if var_name == hash_expr or hash_expr.startswith("tpec_"):
                derived_vars.update(self._re_evaluate_expr(var_name))
        elif holder:
            derived_vars.update(self._evaluate_holders(hash_expr))
        # if the variable has been evaluated then skip updating to prevent infinite loop
        var_modified = self.__is_var_modified_in_context(hash_expr, derived_vars)
        if not var_modified:
            self.__call_on_change(
                var_name,
                value.get() if isinstance(value, _TaipyBase) else value._dict if isinstance(value, _MapDict) else value,
                on_change,
            )
        derived_modified = self.__clean_vars_on_exit()
        if derived_modified is not None:
            self.__send_var_list_update(list(derived_modified), var_name)

    def __call_on_change(self, var_name: str, value: t.Any, on_change: t.Optional[str] = None):
        # TODO: what if _update_function changes 'var_name'... infinite loop?
        var_name_decode, module_name = _variable_decode(self._get_expr_from_hash(var_name))
        current_context = self._get_locals_context()
        if module_name == current_context:
            var_name = var_name_decode
        else:
            if var_name not in self.__var_dir._var_head:
                warnings.warn(f"Can't find matching variable for {var_name} on {current_context} context")
                return
            _found = False
            for k, v in self.__var_dir._var_head[var_name]:
                if v == current_context:
                    var_name = k
                    _found = True
                    break
            if not _found:
                warnings.warn(f"Can't find matching variable for {var_name} on {current_context} context")
                return
        on_change_fn = self._get_user_function(on_change) if on_change else None
        if not callable(on_change_fn):
            on_change_fn = self._get_user_function("on_change")
        if callable(on_change_fn):
            try:
                argcount = on_change_fn.__code__.co_argcount
                args: t.List[t.Any] = [None for _ in range(argcount)]
                if argcount > 0:
                    args[0] = self.__get_state()
                if argcount > 1:
                    args[1] = var_name
                if argcount > 2:
                    args[2] = value
                if argcount > 3:
                    args[3] = current_context
                on_change_fn(*args)
            except Exception as e:
                warnings.warn(f"{on_change or 'on_change'}: callback function raised an exception: {e}")

    def _get_content(self, var_name: str, value: t.Any, image: bool) -> t.Any:
        ret_value = self.__get_content_accessor().get_info(var_name, value, image)
        return Gui.__CONTENT_ROOT + ret_value[0] if isinstance(ret_value, tuple) else ret_value

    def __serve_content(self, path: str) -> t.Any:
        self.__set_client_id_in_context()
        parts = path.split("/")
        if len(parts) > 1:
            file_name = parts[-1]
            (dir_path, as_attachment) = self.__get_content_accessor().get_content_path(
                path[: -len(file_name) - 1], file_name, request.args.get("bypass")
            )
            if dir_path:
                return send_from_directory(str(dir_path), file_name, as_attachment=as_attachment)
        return ("", 404)

    def __upload_files(self):
        self.__set_client_id_in_context()
        if "var_name" not in request.form:
            warnings.warn("No var name")
            return ("No var name", 400)
        var_name = request.form["var_name"]
        multiple = "multiple" in request.form and request.form["multiple"] == "True"
        if "blob" not in request.files:
            warnings.warn("No file part")
            return ("No file part", 400)
        file = request.files["blob"]
        # If the user does not select a file, the browser submits an
        # empty file without a filename.
        if file.filename == "":
            warnings.warn("No selected file")
            return ("No selected file", 400)
        suffix = ""
        complete = True
        part = 0
        if "total" in request.form:
            total = int(request.form["total"])
            if total > 1 and "part" in request.form:
                part = int(request.form["part"])
                suffix = f".part.{part}"
                complete = part == total - 1
        if file:  # and allowed_file(file.filename)
            upload_path = pathlib.Path(self._get_config("upload_folder", tempfile.gettempdir())).resolve()
            file_path = _get_non_existent_file_path(upload_path, secure_filename(file.filename))
            file.save(str(upload_path / (file_path.name + suffix)))
            if complete:
                if part > 0:
                    try:
                        with open(file_path, "wb") as grouped_file:
                            for nb in range(part + 1):
                                with open(upload_path / f"{file_path.name}.part.{nb}", "rb") as part_file:
                                    grouped_file.write(part_file.read())
                    except EnvironmentError as ee:
                        warnings.warn(f"cannot group file after chunk upload {ee}")
                        return
                # notify the file is uploaded
                newvalue = str(file_path)
                if multiple:
                    value = _getscopeattr(self, var_name)
                    if not isinstance(value, t.List):
                        value = [] if value is None else [value]
                    value.append(newvalue)
                    newvalue = value
                setattr(self._bindings(), var_name, newvalue)
        return ("", 200)

    def __send_var_list_update(
        self,
        modified_vars: t.List[str],
        front_var: t.Optional[str] = None,
    ):
        ws_dict = {}
        values = {v: _getscopeattr_drill(self, v) for v in modified_vars}
        for v in values.values():
            if isinstance(v, _TaipyData) and v.get_name() in modified_vars:
                modified_vars.remove(v.get_name())
        for _var in modified_vars:
            newvalue = values.get(_var)
            # self._scopes.broadcast_data(_var, newvalue)
            if isinstance(newvalue, _TaipyData):
                newvalue = None
            else:
                if isinstance(newvalue, (_TaipyContent, _TaipyContentImage)):
                    ret_value = self.__get_content_accessor().get_info(
                        front_var, newvalue.get(), isinstance(newvalue, _TaipyContentImage)
                    )
                    if isinstance(ret_value, tuple):
                        newvalue = Gui.__CONTENT_ROOT + ret_value[0]
                    else:
                        newvalue = ret_value
                elif isinstance(newvalue, _TaipyLov):
                    newvalue = [self.__adapter._run_for_var(newvalue.get_name(), elt) for elt in newvalue.get()]
                elif isinstance(newvalue, _TaipyLovValue):
                    if isinstance(newvalue.get(), list):
                        newvalue = [
                            self.__adapter._run_for_var(newvalue.get_name(), elt, id_only=True)
                            for elt in newvalue.get()
                        ]
                    else:
                        newvalue = self.__adapter._run_for_var(newvalue.get_name(), newvalue.get(), id_only=True)
                if isinstance(newvalue, (dict, _MapDict)):
                    continue  # this var has no transformer
                with warnings.catch_warnings(record=True) as w:
                    warnings.resetwarnings()
                    json.dumps(newvalue, cls=_TaipyJsonEncoder)
                    if len(w):
                        # do not send data that is not serializable
                        continue
            ws_dict[_var] = newvalue
        # TODO: What if value == newvalue?
        self.__send_ws_update_with_dict(ws_dict)

    def __request_data_update(self, var_name: str, payload: t.Any) -> None:
        # Use custom attrgetter function to allow value binding for _MapDict
        newvalue = _getscopeattr_drill(self, var_name)
        if isinstance(newvalue, _TaipyData):
            ret_payload = self._accessors._get_data(self, var_name, newvalue, payload)
            self.__send_ws_update_with_dict({var_name: ret_payload})

    def __request_var_update(self, payload: t.Any):
        if isinstance(payload, dict) and isinstance(payload.get("names"), list):
            self.__send_var_list_update(payload["names"])

    def __send_ws(self, payload: dict) -> None:
        grouping_message = self.__get_message_grouping()
        if grouping_message is None:
            try:
                self._server._ws.emit(
                    "message",
                    payload,
                    to=self.__get_ws_receiver(),
                )
            except Exception as e:
                warnings.warn(f"Web Socket communication error in {self.__frame.f_code.co_name}\n{e}")
        else:
            grouping_message.append(payload)

    def _send_ws_id(self, id: str) -> None:
        self.__send_ws(
            {
                "type": _WsType.CLIENT_ID.value,
                "id": id,
            }
        )

    def _send_ws_download(self, content: str, name: str, on_action: str) -> None:
        self.__send_ws({"type": _WsType.DOWNLOAD_FILE.value, "content": content, "name": name, "on_action": on_action})

    def __send_ws_alert(self, type: str, message: str, system_notification: bool, duration: int) -> None:
        self.__send_ws(
            {
                "type": _WsType.ALERT.value,
                "atype": type,
                "message": message,
                "system": system_notification,
                "duration": duration,
            }
        )

    def __send_ws_partial(self, partial: str):
        self.__send_ws(
            {
                "type": _WsType.PARTIAL.value,
                "name": partial,
            }
        )

    def __send_ws_block(
        self,
        action: t.Optional[str] = None,
        message: t.Optional[str] = None,
        close: t.Optional[bool] = False,
        cancel: t.Optional[bool] = False,
    ):
        self.__send_ws(
            {
                "type": _WsType.BLOCK.value,
                "action": action,
                "close": close,
                "message": message,
                "noCancel": not cancel,
            }
        )

    def __send_ws_navigate(
        self,
        to: t.Optional[str] = None,
    ):
        self.__send_ws(
            {
                "type": _WsType.NAVIGATE.value,
                "to": to,
            }
        )

    def __send_ws_update_with_dict(self, modified_values: dict) -> None:
        payload = [
            {"name": _get_client_var_name(k), "payload": (v if isinstance(v, dict) and "value" in v else {"value": v})}
            for k, v in modified_values.items()
        ]
        self.__send_ws({"type": _WsType.MULTIPLE_UPDATE.value, "payload": payload})

    def __get_ws_receiver(self) -> t.Union[t.List[str], t.Any, None]:
        if self._bindings()._get_single_client():
            return None
        sid = getattr(request, "sid", None) if request else None
        sids = self.__client_id_2_sid.get(self._get_client_id(), set())
        if sid:
            sids.add(sid)
        return list(sids)

    def __get_message_grouping(self):
        if not _hasscopeattr(self, Gui.__MESSAGE_GROUPING_NAME):
            return None
        return _getscopeattr(self, Gui.__MESSAGE_GROUPING_NAME)

    def __enter__(self):
        self.__hold_messages()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self.__send_messages()
        except Exception as e:
            warnings.warn(f"An exception was raised while sending messages: {e}")
        if exc_value:
            warnings.warn(f"An {exc_type or 'Exception'} was raised: {exc_value}")
        return True

    def __hold_messages(self):
        grouping_message = self.__get_message_grouping()
        if grouping_message is None:
            self._bind_var_val(Gui.__MESSAGE_GROUPING_NAME, [])

    def __send_messages(self):
        grouping_message = self.__get_message_grouping()
        if grouping_message is not None:
            _delscopeattr(self, Gui.__MESSAGE_GROUPING_NAME)
            if len(grouping_message):
                self.__send_ws({"type": _WsType.MULTIPLE_MESSAGE.value, "payload": grouping_message})

    def _get_user_function(self, func_name: str) -> t.Union[t.Callable, str]:
        func = _getscopeattr(self, func_name, None)
        if not callable(func):
            func = self._get_locals_bind().get(func_name)
        if not callable(func):
            func = self.__locals_context.get_default().get(func_name)
        if callable(func):
            return func
        return func_name

    def __on_action(self, id: t.Optional[str], payload: t.Any) -> None:
        action = payload.get("action") if isinstance(payload, dict) else str(payload)
        if action:
            if self.__call_function_with_args(
                action_function=self._get_user_function(action), id=id, payload=payload, action=action
            ):
                return
            else:
                warnings.warn(f"on_action: '{action}' is not a function")
        if hasattr(self, "on_action"):
            self.__call_function_with_args(action_function=self.on_action, id=id, payload=payload, action=action)

    def __call_function_with_args(self, **kwargs):
        action_function = kwargs.get("action_function")
        id = kwargs.get("id")
        action = kwargs.get("action")
        payload = kwargs.get("payload")

        if callable(action_function):
            try:
                argcount = action_function.__code__.co_argcount
                args = [None for _ in range(argcount)]
                if argcount > 0:
                    args[0] = self.__get_state()
                if argcount > 1:
                    args[1] = id
                if argcount > 2:
                    args[2] = payload if action is None else action
                if argcount > 3 and action is not None:
                    args[3] = payload
                action_function(*args)
                return True
            except Exception as e:
                warnings.warn(f"on_action: '{action_function.__name__}' function invocation exception: {e}")
        return False

    def _call_function_with_state(self, user_function: t.Callable, args: t.List[t.Any]) -> t.Any:
        args.insert(0, self.__get_state())
        arg_count = user_function.__code__.co_argcount
        if arg_count > len(args):
            args += (arg_count - len(args)) * [None]
        else:
            args = args[:arg_count]
        return user_function(*args)

    def _call_user_callback(self, context_id: t.Optional[str], user_callback: t.Callable, args: t.List[t.Any]) -> t.Any:
        try:
            with self.get_flask_app().app_context():
                self.__set_client_id_in_context(context_id)
                return self._call_function_with_state(user_callback, args)
        except Exception as e:
            warnings.warn(f"invoke_state_callback: '{user_callback.__name__}' function invocation exception: {e}")
        return None

    # Proxy methods for Evaluator
    def _evaluate_expr(self, expr: str) -> t.Any:
        return self.__evaluator.evaluate_expr(self, expr)

    def _re_evaluate_expr(self, var_name: str) -> t.Set[str]:
        return self.__evaluator.re_evaluate_expr(self, var_name)

    def _get_expr_from_hash(self, hash: str) -> str:
        return self.__evaluator.get_expr_from_hash(hash)

    def _evaluate_bind_holder(self, holder: t.Type[_TaipyBase], expr: str) -> str:
        return self.__evaluator.evaluate_bind_holder(self, holder, expr)

    def _evaluate_holders(self, expr: str) -> t.List[str]:
        return self.__evaluator.evaluate_holders(self, expr)

    def _is_expression(self, expr: str) -> bool:
        return self.__evaluator._is_expression(expr)

    # Proxy methods for Adapter
    def _add_adapter_for_type(self, type_name: str, adapter: t.Callable) -> None:
        self.__adapter._add_for_type(type_name, adapter)

    def _add_type_for_var(self, var_name: str, type_name: str) -> None:
        self.__adapter._add_type_for_var(var_name, type_name)

    def _get_adapter_for_type(self, type_name: str) -> t.Optional[t.Callable]:
        return self.__adapter._get_for_type(type_name)

    def _get_unique_type_adapter(self, type_name: str) -> str:
        return self.__adapter._get_unique_type(type_name)

    def _run_adapter(
        self, adapter: t.Optional[t.Callable], value: t.Any, var_name: str, id_only=False
    ) -> t.Union[t.Tuple[str, ...], str, None]:
        return self.__adapter._run(adapter, value, var_name, id_only)

    def _get_valid_adapter_result(self, value: t.Any, id_only=False) -> t.Union[t.Tuple[str, ...], str, None]:
        return self.__adapter._get_valid_result(value, id_only)

    def _is_ui_blocked(self):
        return _getscopeattr(self, Gui.__UI_BLOCK_NAME, False)

    def __get_on_cancel_block_ui(self, callback: t.Optional[str]):
        def _taipy_on_cancel_block_ui(guiApp, id: t.Optional[str], payload: t.Any):
            if _hasscopeattr(self, Gui.__UI_BLOCK_NAME):
                _setscopeattr(self, Gui.__UI_BLOCK_NAME, False)
            self.__on_action(id, callback)

        return _taipy_on_cancel_block_ui

    def __add_pages_in_folder(self, folder_name: str, folder_path: str):
        list_of_files = os.listdir(folder_path)
        for file_name in list_of_files:
            from .renderers import Html, Markdown

            if re_match := Gui.__RE_HTML.match(file_name):
                renderers = Html(os.path.join(folder_path, file_name), frame=None)
                renderers.modify_taipy_base_url(folder_name)
                self.add_page(name=f"{folder_name}/{re_match.group(1)}", page=renderers)
            elif re_match := Gui.__RE_MD.match(file_name):
                renderers_md = Markdown(os.path.join(folder_path, file_name), frame=None)
                self.add_page(name=f"{folder_name}/{re_match.group(1)}", page=renderers_md)
            elif os.path.isdir(child_dir_path := os.path.join(folder_path, file_name)):
                child_dir_name = f"{folder_name}/{file_name}"
                self.__add_pages_in_folder(child_dir_name, child_dir_path)

    # Proxy methods for LocalsContext
    def _get_locals_bind(self) -> t.Dict[str, t.Any]:
        return self.__locals_context.get_locals()

    def _get_locals_context(self) -> str:
        current_context = self.__locals_context.get_context()
        return current_context if current_context is not None else self.__default_module_name

    def _set_locals_context(self, context: t.Optional[str]) -> None:
        self.__locals_context.set_locals_context(context)

    def _reset_locals_context(self) -> None:
        self.__locals_context.reset_locals_context()

    @staticmethod
    def _get_root_page_name():
        return Gui.__root_page_name

    def _set_flask(self, flask: Flask):
        self._flask = flask

    def _get_default_module_name(self):
        return self.__default_module_name

    @staticmethod
    def _get_timezone() -> str:
        return Gui.__LOCAL_TZ

    @staticmethod
    def _set_timezone(tz: str):
        Gui.__LOCAL_TZ = tz

    # Public methods
    def add_page(
        self,
        name: str,
        page: t.Union[str, Page],
        style: t.Optional[str] = "",
    ) -> None:
        """Add a page to the graphical User Interface.

        Arguments:
            name: The name of the page.
            page (Union[str, Page^]): The content of the page.<br/>
                It can be an instance of `Markdown^` or `Html^`.<br/>
                If _page_ is a string, then:

                - If _page_ is set to the pathname of a readable file, the page
                  content is read as Markdown input text.
                - If it is not, the page content is read from this string as
                  Markdown text.
            style (Optional[str]): Additional CSS style to apply to this page.

                - if there is style associated with a page, it is used at a global level
                - if there is no style associated with the page, the style is cleared at a global level
                - if the page is embedded in a block control, the style is ignored

        Note that page names cannot start with the slash ('/') character and that each
        page must have a unique name.
        """
        # Validate name
        if name is None:
            raise Exception("name is required for add_page function!")
        if not Gui.__RE_PAGE_NAME.match(name):
            raise SyntaxError(
                f'Page name "{name}" is invalid. It must only contain letters, digits, dash (-), underscore (_), and forward slash (/) characters.'
            )
        if name.startswith("/"):
            raise SyntaxError(f'Page name "{name}" cannot start with forward slash (/) character')
        if name in self._config.routes:
            raise Exception(f'Page name "{name if name != Gui.__root_page_name else "/"}" is already defined')
        if isinstance(page, str):
            from .renderers import Markdown

            page = Markdown(page, frame=None)
        elif not isinstance(page, Page):
            raise Exception(f'"page" is invalid for page name "{name if name != Gui.__root_page_name else "/"}')
        # Init a new page
        new_page = _Page()
        new_page._route = name
        new_page._renderer = page
        new_page._style = style
        # Append page to _config
        self._config.pages.append(new_page)
        self._config.routes.append(name)
        # Update locals context
        self.__locals_context.add(page._get_module_name(), page._get_locals())
        # Update variable directory
        self.__var_dir.add_frame(page._frame)

    def add_pages(self, pages: t.Union[t.Dict[str, t.Union[str, Page]], str] = None) -> None:
        """Add several pages to the graphical User Interface.

        Arguments:
            pages (Union[dict[str, Union[str, Page^]], str]): The pages to add.<br/>
                If _pages_ is a dictionnary, a page is added to this `Gui` instance
                for each of the entries in _pages_:

                - The entry key is used as the page name.
                - The entry value is used as the page content:
                    - The value can can be an instance of `Markdown^` or `Html^`, then
                      it is used as the page definition.
                    - If entry value is a string, then:
                        - If it is set to the pathname of a readable file, the page
                          content is read as Markdown input text.
                        - If it is not, the page content is read from this string as
                          Markdown text.

                If _pages_ is a string that contains the path to a directory, then
                this directory is traversed, looking for filenames that have the
                _.md_ extention, TODO


        !!! note "Reading pages from a directory"
            If _pages_ is a string that holds the path to a readable directory, then
            this directory is traversed, recursively, to find files that Taipy can build
            pages from.

            For every new directory that is traversed, a new hierarchical level
            for pages is created.

            For every file that is found:

            - If the filename extention is _.md_, it is read as Markdown content and
             a new page is created with the base name of this filename.
            - If the filename extention is _.html_, it is read as HTML content and
             a new page is created with the base name of this filename.

            For example, say you have the following directory structure:
            ```
            reports
            ├── home.html
            ├── budget/
            │   ├── expenses/
            │   │   ├── marketing.md
            │   │   └── production.md
            │   └── revenue/
            │       ├── EMAE.md
            │       ├── USA.md
            │       └── ASIA.md
            └── cashflow/
                ├── weekly.md
                ├── monthly.md
                └── yearly.md
            ```

            Calling `gui.add_pages('reports')` is equivalent to calling:
            ```py
            gui.add_pages({
                            "reports/home", Html("reports/home.html"),
                            "reports/budget/expenses/marketing", Markdown("reports/budget/expenses/marketing.md"),
                            "reports/budget/expenses/production", Markdown("reports/budget/expenses/production.md"),
                            "reports/budget/revenue/EMAE", Markdown("reports/budget/revenue/EMAE.md"),
                            "reports/budget/revenue/USA", Markdown("reports/budget/revenue/USA.md"),
                            "reports/budget/revenue/ASIA", Markdown("reports/budget/revenue/ASIA.md"),
                            "reports/cashflow/weekly", Markdown("reports/cashflow/weekly.md"),
                            "reports/cashflow/monthly", Markdown("reports/cashflow/monthly.md"),
                            "reports/cashflow/yearly", Markdown("reports/cashflow/yearly.md")
            })
            ```
        """
        if isinstance(pages, dict):
            for k, v in pages.items():
                if k == "/":
                    k = Gui.__root_page_name
                self.add_page(name=k, page=v)
        elif isinstance(folder_name := pages, str):
            if not hasattr(self, "_root_dir"):
                self._root_dir = os.path.dirname(inspect.getabsfile(self.__frame))
            folder_path = folder_name if os.path.isabs(folder_name) else os.path.join(self._root_dir, folder_name)
            folder_name = os.path.basename(folder_path)
            if not os.path.isdir(folder_path):
                raise RuntimeError(f"Path {folder_path} is not a valid directory")
            if folder_name in self.__directory_name_of_pages:
                raise Exception(f"Base directory name {folder_name} of path {folder_path} is not unique")
            if folder_name in Gui.__reserved_routes:
                raise Exception(f"Invalid directory. Directory {folder_name} is a reserved route")
            self.__directory_name_of_pages.append(folder_name)
            self.__add_pages_in_folder(folder_name, folder_path)

    # partials

    def add_partial(
        self,
        page: t.Union[str, Page],
    ) -> Partial:
        """Create a new `Partial^`.

        The [User Manual section on Partials](../../gui/pages/#partials) gives details on
        when and how to use this class.

        Arguments:
            page (Union[str, Page^]): The page to create a new Partial from.<br/>
                It can be an instance of `Markdown^` or `Html^`.<br/>
                If _page_ is a string, then:

                - If _page_ is set to the pathname of a readable file, the content of
                  the new `Partial` is read as Markdown input text.
                - If it is not, the content of the new `Partial` is read from this string
                  as Markdown text.

        Returns:
            Partial: The new `Partial` object defined by _page_.
        """
        new_partial = Partial()
        # Validate name
        if new_partial._route in self._config.partial_routes or new_partial._route in self._config.routes:
            warnings.warn(f'Partial name "{new_partial._route}" is already defined')
        if isinstance(page, str):
            from .renderers import Markdown

            page = Markdown(page, frame=None)
        elif not isinstance(page, Page):
            raise Exception(f'Partial name "{new_partial._route}" has invalid Page')
        new_partial._renderer = page
        # Append partial to _config
        self._config.partials.append(new_partial)
        self._config.partial_routes.append(str(new_partial._route))
        # Update locals context
        self.__locals_context.add(page._get_module_name(), page._get_locals())
        # Update variable directory
        self.__var_dir.add_frame(page._frame)
        return new_partial

    def _update_partial(self, partial: Partial):
        partials = _getscopeattr(self, Partial._PARTIALS, {})
        partials[partial._route] = partial
        _setscopeattr(self, Partial._PARTIALS, partials)
        self.__send_ws_partial(str(partial._route))

    def _get_partial(self, route: str) -> t.Optional[Partial]:
        partials = _getscopeattr(self, Partial._PARTIALS, {})
        partial = partials.get(route)
        if partial is None:
            partial = next((p for p in self._config.partials if p._route == route), None)
        return partial

    # Main binding method (bind in markdown declaration)
    def _bind_var(self, var_name: str) -> str:
        encoded_var_name = self.__var_dir.add_var(var_name, self._get_locals_context())
        if not hasattr(self._bindings(), encoded_var_name) and var_name in self._get_locals_bind().keys():
            self._bind(encoded_var_name, self._get_locals_bind()[var_name])
        return encoded_var_name

    def _bind_var_val(self, var_name: str, value: t.Any) -> bool:
        if not _RE_TPMDL_DECODE.match(var_name):
            var_name = self.__var_dir.add_var(var_name, self._get_locals_context())
        if not hasattr(self._bindings(), var_name):
            self._bind(var_name, value)
            return True
        return False

    def __bind_local_func(self, name: str):
        func = getattr(self, name, None)
        if func is not None and not callable(func):
            warnings.warn(
                f"{self.__class__.__name__}.{name}: {func} should be a function; looking for {name} in the script."
            )
            func = None
        if func is None:
            func = self._get_locals_bind().get(name)
        if func is not None:
            if callable(func):
                setattr(self, name, func)
            else:
                warnings.warn(f"{name}: {func} should be a function")

    def load_config(self, config: Config) -> None:
        self._config._load(config)

    def _download(self, content: t.Any, name: t.Optional[str] = "", on_action: t.Optional[str] = ""):
        content_str = self._get_content("Gui.download", content, False)
        self._send_ws_download(content_str, str(name), str(on_action))

    def _notify(
        self,
        notification_type: str = "I",
        message: str = "",
        system_notification: t.Optional[bool] = None,
        duration: t.Optional[int] = None,
    ):
        self.__send_ws_alert(
            notification_type,
            message,
            self._get_config("system_notification", False) if system_notification is None else system_notification,
            self._get_config("notification_duration", 3000) if duration is None else duration,
        )

    def _hold_actions(
        self,
        callback: t.Optional[t.Union[str, t.Callable]] = None,
        message: t.Optional[str] = "Work in Progress...",
    ):  # pragma: no cover
        action_name = callback.__name__ if callable(callback) else callback
        func = self.__get_on_cancel_block_ui(action_name)
        def_action_name = func.__name__
        _setscopeattr(self, def_action_name, func)

        if _hasscopeattr(self, Gui.__UI_BLOCK_NAME):
            _setscopeattr(self, Gui.__UI_BLOCK_NAME, True)
        else:
            self._bind(Gui.__UI_BLOCK_NAME, True)
        self.__send_ws_block(action=def_action_name, message=message, cancel=bool(action_name))

    def _resume_actions(self):  # pragma: no cover
        if _hasscopeattr(self, Gui.__UI_BLOCK_NAME):
            _setscopeattr(self, Gui.__UI_BLOCK_NAME, False)
        self.__send_ws_block(close=True)

    def _navigate(self, to: t.Optional[str] = ""):
        to = to or Gui.__root_page_name
        if to not in self._config.routes:
            warnings.warn(f'cannot navigate to "{to if to != Gui.__root_page_name else "/"}": unknown page.')
            return
        self.__send_ws_navigate(to)

    def __init_route(self):
        self.__set_client_id_in_context()
        if hasattr(self, "on_init") and callable(self.on_init):
            if not _hasscopeattr(self, Gui.__ON_INIT_NAME):
                _setscopeattr(self, Gui.__ON_INIT_NAME, True)
                try:
                    self._call_function_with_state(self.on_init, [])
                except Exception as e:
                    warnings.warn(f"Exception on on_init execution \n{e}")
        return self._render_route()

    def __render_page(self, page_name: str) -> t.Any:
        self.__set_client_id_in_context()
        page = next((page_i for page_i in self._config.pages if page_i._route == page_name), None)

        # try partials
        if page is None:
            page = self._get_partial(page_name)
        # Make sure that there is a page instance found
        if page is None:
            return (jsonify({"error": "Page doesn't exist!"}), 400, {"Content-Type": "application/json; charset=utf-8"})
        context = page.render(self)
        if (
            page_name == Gui.__root_page_name
            and page._rendered_jsx is not None
            and "<PageContent" not in page._rendered_jsx
        ):
            page._rendered_jsx += "<PageContent />"
        # Return jsx page
        if page._rendered_jsx is not None:
            return self._server._render(
                page._rendered_jsx, page._style if page._style is not None else "", page._head, context
            )
        else:
            return ("No page template", 404)

    def _render_route(self) -> t.Any:
        router = '<Routes key="routes">'
        router += (
            '<Route path="/" key="'
            + Gui.__root_page_name
            + '" element={<MainPage key="tr'
            + Gui.__root_page_name
            + '" path="/'
            + Gui.__root_page_name
            + '"'
        )
        routes = self._config.routes
        route = next((r for r in routes if r != Gui.__root_page_name), None)
        router += (' route="/' + route + '"') if route else ""
        router += " />} >"
        locations = {"/": f"/{Gui.__root_page_name}"}
        for route in routes:
            if route != Gui.__root_page_name:
                router += (
                    '<Route path="'
                    + route
                    + '" key="'
                    + route
                    + '" element={<TaipyRendered key="tr'
                    + route
                    + '"/>} />'
                )
                locations[f"/{route}"] = f"/{route}"
        router += '<Route path="*" key="NotFound" element={<NotFound404 />} />'
        router += "</Route>"
        router += "</Routes>"

        return self._server._direct_render_json(
            {
                "router": router,
                "locations": locations,
                "blockUI": self._is_ui_blocked(),
            }
        )

    def _register_data_accessor(self, data_accessor_class: t.Type[_DataAccessor]) -> None:
        self._accessors._register(data_accessor_class)

    def get_flask_app(self) -> Flask:
        """Get the internal Flask application.

        Returns:
            the Flask instance used.
        """
        return self._server.get_flask()

    def _set_frame(self, frame: FrameType):
        if not isinstance(frame, FrameType):
            raise RuntimeError("frame must be a FrameType where Gui can collect the local variables.")
        self.__frame = frame
        self.__default_module_name = _get_module_name_from_frame(self.__frame)

    def _set_state(self, state: State):
        if isinstance(state, State):
            self.__state = state

    def run(
        self,
        run_server: bool = True,
        run_in_thread: bool = False,
        ssl_context: t.Optional[t.Union[ssl.SSLContext, t.Tuple[str, t.Optional[str]], t.Literal["adhoc"]]] = None,
        **kwargs,
    ) -> t.Optional[Flask]:
        """
        Starts the server that delivers pages to Web clients.

        Once you enter `run()`, users can run Web browsers and point to the Web server
        URL that `Gui` serves. The default is to listen to the _localhost_ address
        (127.0.0.1) on the port number 5000. However, the configuration of this `Gui`
        object may impact that (see the
        [Configuration](../gui/configuration.md#configuring-the-gui-instance)
        section of the User Manual for details).

        Arguments:
            run_server (bool): Whether or not to run a Web server locally.
                If set to _False_, a Web server is _not_ created and started.
            run_in_thread (bool): Whether or not to run a Web server in a separated thread.
                If set to _True_, the Web server is run is a separated thread.
                Note that if you are running in an IPython notebook context, the Web
                server is always run in a separate thread.
            ssl_context (Optional[Union[ssl.SSLContext, Tuple[str, Optional[str]], te.Literal['adhoc']]]):
                Configure TLS to serve over HTTPS. Can be an ssl.SSLContext object, a (cert_file, key_file) tuple to
                create a typical context, or the string 'adhoc' to generate a temporary self-signed certificate.</br>
                The default value is None.
            **kwargs: Additional keyword arguments that configure how this `Gui` is run.
                Please refer to the
                [Configuration](../gui/configuration.md#configuring-the-gui-instance)
                section in the User Manual for more information.

        Returns:
            The Flask instance if _run_server_ is _False_ else _None_.
        """

        app_config = self._config.config

        run_root_dir = os.path.dirname(inspect.getabsfile(self.__frame))

        # Register _root_dir for abs path
        if not hasattr(self, "_root_dir"):
            self._root_dir = run_root_dir

        # Load application config from multiple sources (env files, kwargs, command line)
        self._config._build_config(run_root_dir, self.__env_filename, kwargs)

        # Init server if there is no server
        if not hasattr(self, "_server"):
            self._server = _Server(
                self,
                path_mapping=self._path_mapping,
                flask=self._flask,
                css_file=self._css_file,
                content_security_policy=self._get_config("content_security_policy", None),
                force_https=self._get_config("force_https", False),
            )

        # Stop and reinitialize the server if it is still running as a thread
        if (_is_in_notebook() or run_in_thread) and hasattr(self._server, "_thread"):
            self._server._thread.kill()
            self._server._thread.join()
            self._flask_blueprint = []
            self._server = _Server(
                self,
                path_mapping=self._path_mapping,
                flask=self._flask,
                css_file=self._css_file,
                content_security_policy=self._get_config("content_security_policy", None),
                force_https=self._get_config("force_https", False),
            )
            self._bindings()._new_scopes()

        # Special config for notebook runtime
        if _is_in_notebook() or run_in_thread:
            self._config.config["single_client"] = True

        if run_server and app_config["ngrok_token"]:  # pragma: no cover
            if not util.find_spec("pyngrok"):
                raise RuntimeError("Cannot use ngrok as pyngrok package is not installed")
            ngrok.set_auth_token(app_config["ngrok_token"])
            http_tunnel = ngrok.connect(app_config["port"], "http")
            app_config["use_reloader"] = False
            print(f" * NGROK Public Url: {http_tunnel.public_url}")

        locals_bind = _filter_locals(self.__frame.f_locals)

        self.__locals_context.set_default(locals_bind)

        self.__var_dir.set_default(self.__frame)

        if self.__state is None:
            self.__state = State(self, self.__locals_context.get_all_keys())

        if _is_in_notebook():
            # to allow gui.state.x in notebook mode
            self.state = self.__state

        with self.get_flask_app().app_context():
            self.__var_dir.process_imported_var()
            # bind on_change and on_action function if available
            self.__bind_local_func("on_init")
            self.__bind_local_func("on_change")
            self.__bind_local_func("on_action")

        # base global ctx is TaipyHolder classes + script modules and callables
        glob_ctx = {t.__name__: t for t in _TaipyBase.__subclasses__()}
        glob_ctx.update({k: v for k, v in locals_bind.items() if inspect.ismodule(v) or callable(v)})
        self.__evaluator = _Evaluator(glob_ctx)

        # add en empty main page if it is not defined
        if Gui.__root_page_name not in self._config.routes:
            new_page = _Page()
            new_page._route = Gui.__root_page_name
            new_page._renderer = _EmptyPage()
            self._config.pages.append(new_page)
            self._config.routes.append(Gui.__root_page_name)

        pages_bp = Blueprint("taipy_pages", __name__)
        self._flask_blueprint.append(pages_bp)

        # server URL Rule for taipy images
        images_bp = Blueprint("taipy_images", __name__)
        images_bp.add_url_rule(f"{Gui.__CONTENT_ROOT}<path:path>", view_func=self.__serve_content)

        self._flask_blueprint.append(images_bp)

        # server URL for uploaded files
        upload_bp = Blueprint("taipy_upload", __name__)
        upload_bp.add_url_rule(Gui.__UPLOAD_URL, view_func=self.__upload_files, methods=["POST"])
        self._flask_blueprint.append(upload_bp)

        _absolute_path = str(pathlib.Path(__file__).parent.resolve())
        self._flask_blueprint.append(
            self._server._get_default_blueprint(
                static_folder=f"{_absolute_path}{os.path.sep}webapp",
                template_folder=f"{_absolute_path}{os.path.sep}webapp",
                title=self._get_config("title", "Taipy App"),
                favicon=self._get_config("favicon", "/favicon.png"),
                root_margin=self._get_config("margin", None),
            )
        )

        # Run parse markdown to force variables binding at runtime
        pages_bp.add_url_rule("/taipy-jsx/<path:page_name>", view_func=self.__render_page)

        # server URL Rule for flask rendered react-router
        pages_bp.add_url_rule("/taipy-init", view_func=self.__init_route)

        # Register Flask Blueprint if available
        for bp in self._flask_blueprint:
            self._server.get_flask().register_blueprint(bp)

        # Register data accessor communicaiton data format (JSON, Apache Arrow)
        self._accessors._set_data_format(_DataFormat.APACHE_ARROW if app_config["use_arrow"] else _DataFormat.JSON)

        # Use multi user or not
        self._bindings()._set_single_client(bool(app_config["single_client"]))

        # Start Flask Server
        if not run_server:
            return self.get_flask_app()

        return self._server.runWithWS(
            host=app_config["host"],
            port=app_config["port"],
            debug=app_config["debug"],
            use_reloader=app_config["use_reloader"],
            flask_log=app_config["flask_log"],
            run_in_thread=run_in_thread,
            ssl_context=ssl_context,
        )

    def stop(self):
        """
        Stops the Web server.

        This function stops the underlying Web server only in the situation where
        it was run in a separated thread: the _run_in_thread_ parameter to the
        `(Gui.)run^` method was set to True, or you are running in an IPython notebook
        context.
        """
        if hasattr(self, "_server") and hasattr(self._server, "_thread"):
            self._server._thread.kill()
            self._server._thread.join()
            print("Gui server has been stopped")



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import typing as t
from importlib import util

import numpy as np

_has_rdp_module = False

if util.find_spec("rdp"):
    from rdp import rdp

    _has_rdp_module = True


if t.TYPE_CHECKING:
    import pandas as pd


def _df_data_filter(
    dataframe: pd.DataFrame,
    x_column_name: t.Union[None, str],
    y_column_name: str,
    expected_row_count: t.Union[None, int],
    margin_of_error: float = 0.3,
) -> t.Tuple[pd.DataFrame, t.Union[None, str], str]:
    if not _has_rdp_module or (
        expected_row_count is not None and __df_row_count(dataframe) <= expected_row_count * (1 + margin_of_error)
    ):
        return dataframe, x_column_name, y_column_name
    df = dataframe.copy()
    if not x_column_name:
        index = 0
        while f"tAiPy_index_{index}" in df.columns:
            index += 1
        x_column_name = f"tAiPy_index_{index}"
        df[x_column_name] = df.index
    if expected_row_count is None:
        expected_row_count = int(__df_row_count(df) * 0.1)
    points = df[[x_column_name, y_column_name]].to_numpy()
    return df[__np_filter(points, int(expected_row_count), margin_of_error)], x_column_name, y_column_name


def __np_filter(points: np.ndarray, expected_row_count: int, margin_of_error: float) -> np.ndarray:
    # Perform binary search to find the appropriate epsilon for rdp algorithm
    # epsilon will determine number of rows in dataframe
    upper_epsilon = round(np.diff(points[:, 1]).max())
    lower_epsilon = 0
    epsilon = 50  # good starting point for epsilon
    upper_count = (1 + margin_of_error) * expected_row_count
    lower_count = (1 - margin_of_error) * expected_row_count
    result = rdp(points, epsilon=epsilon, return_mask=True)
    count = np.count_nonzero(result)
    # The search will end if the number of rows fits within this `count` range or `upper_epsilon` is smaller than `lower_epsilon`
    while upper_epsilon > lower_epsilon and (count > upper_count or count < lower_count):
        if count > expected_row_count:
            lower_epsilon = epsilon + 1
        else:
            upper_epsilon = epsilon - 1
        epsilon = round((upper_epsilon + lower_epsilon) / 2)
        result = rdp(points, epsilon=epsilon, return_mask=True)
        count = np.count_nonzero(result)
    return result


def __df_row_count(df: pd.DataFrame) -> int:
    return df.shape[0]



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import re
import typing as t
from types import FrameType

from ._locals_context import _LocalsContext
from .get_imported_var import _get_imported_var
from .get_module_name import _get_module_name_from_frame, _get_module_name_from_imported_var


class _VariableDirectory:
    def __init__(self, locals_context: _LocalsContext):
        self._locals_context = locals_context
        self._default_module = ""
        self._var_dir: t.Dict[str, t.Dict] = {}
        self._var_head: t.Dict[str, t.List[t.Tuple[str, str]]] = {}
        self._imported_var_dir: t.Dict[str, t.List[t.Tuple[str, str, str]]] = {}

    def set_default(self, frame: FrameType) -> None:
        self._default_module = _get_module_name_from_frame(frame)
        self.add_frame(frame)

    def add_frame(self, frame: t.Optional[FrameType]) -> None:
        if frame is None:
            return
        module_name = _get_module_name_from_frame(frame)
        if module_name not in self._imported_var_dir:
            imported_var_list = _get_imported_var(frame)
            self._imported_var_dir[module_name] = imported_var_list

    def process_imported_var(self) -> None:
        default_imported_dir = self._imported_var_dir[self._default_module]
        self._locals_context.set_locals_context(self._default_module)
        for name, asname, module in default_imported_dir:
            imported_module_name = _get_module_name_from_imported_var(
                name, self._locals_context.get_locals()[asname], module
            )
            temp_var_name = self.add_var(asname, self._default_module)
            self.add_var(name, imported_module_name, temp_var_name)
        self._locals_context.reset_locals_context()

        for k, v in self._imported_var_dir.items():
            self._locals_context.set_locals_context(k)
            for name, asname, module in v:
                imported_module_name = _get_module_name_from_imported_var(
                    name, self._locals_context.get_locals()[asname], module
                )
                var_name = self.get_var(name, imported_module_name)
                var_asname = self.get_var(asname, k)
                if var_name is None and var_asname is None:
                    temp_var_name = self.add_var(asname, k)
                    self.add_var(name, imported_module_name, temp_var_name)
                elif var_name is not None:
                    self.add_var(asname, k, var_name)
                else:
                    self.add_var(name, imported_module_name, var_asname)
            self._locals_context.reset_locals_context()

    def add_var(self, name: str, module: t.Optional[str], var_name: t.Optional[str] = None) -> str:
        if module is None:
            module = self._default_module
        if gv := self.get_var(name, module):
            return gv
        var_encode = _variable_encode(name, module) if module != self._default_module else name
        if var_name is None:
            var_name = var_encode
        self.__add_var_head(name, module, var_name)
        if var_encode != var_name:
            var_name_decode, module_decode = _variable_decode(var_name)
            if module_decode is None:
                module_decode = self._default_module
            self.__add_var_head(var_name_decode, module_decode, var_encode)
        if name not in self._var_dir:
            self._var_dir[name] = {module: var_name}
        else:
            self._var_dir[name][module] = var_name
        return var_name

    def __add_var_head(self, name: str, module: str, var_head: str) -> None:
        if var_head not in self._var_head:
            self._var_head[var_head] = [(name, module)]
        else:
            self._var_head[var_head].append((name, module))

    def get_var(self, name: str, module: str) -> t.Optional[str]:
        if name in self._var_dir and module in self._var_dir[name]:
            return self._var_dir[name][module]
        return None


_MODULE_NAME_MAP: t.List[str] = []
_RE_TPMDL_DECODE = re.compile(r"(.*?)_TPMDL_(.*)")


def _variable_encode(var_name: str, module_name: t.Optional[str]):
    if module_name is None:
        return var_name
    if module_name not in _MODULE_NAME_MAP:
        _MODULE_NAME_MAP.append(module_name)
    return f"{var_name}_TPMDL_{_MODULE_NAME_MAP.index(module_name)}"


def _variable_decode(var_name: str):
    from ._evaluator import _Evaluator

    if result := _RE_TPMDL_DECODE.match(var_name):
        return _Evaluator._expr_decode(str(result[1])), _MODULE_NAME_MAP[int(result[2])]
    return _Evaluator._expr_decode(var_name), None


def _reset_name_map():
    _MODULE_NAME_MAP.clear()



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t

from ..factory import _Factory


class _HtmlFactory(_Factory):
    @staticmethod
    def create_element(gui, control_type: str, all_properties: t.Dict[str, str]) -> t.Tuple[str, str]:
        builder = _Factory.CONTROL_BUILDERS[control_type](gui, control_type, all_properties)
        if not builder:
            return f"<div>INVALID SYNTAX - Control is '{control_type}'", "div"
        builder_str, element_name = builder.build_to_string()
        return builder_str, element_name



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t
from enum import Enum

from .utils import (
    _TaipyBase,
    _TaipyBool,
    _TaipyContent,
    _TaipyContentImage,
    _TaipyData,
    _TaipyDate,
    _TaipyLov,
    _TaipyLovValue,
    _TaipyNumber,
)


class _WsType(Enum):
    ACTION = "A"
    MULTIPLE_UPDATE = "MU"
    REQUEST_UPDATE = "RU"
    DATA_UPDATE = "DU"
    UPDATE = "U"
    ALERT = "AL"
    BLOCK = "BL"
    NAVIGATE = "NA"
    CLIENT_ID = "ID"
    MULTIPLE_MESSAGE = "MS"
    DOWNLOAD_FILE = "DF"
    PARTIAL = "PR"


NumberTypes = {"int", "int64", "float", "float64"}


class _AttributeType(Enum):
    boolean = "boolean"
    content = _TaipyContent
    data = _TaipyData
    date = _TaipyDate
    dict = "dict"
    dynamic_number = _TaipyNumber
    dynamic_boolean = _TaipyBool
    dynamic_list = "dynamiclist"
    dynamic_string = "dynamicstring"
    function = "function"
    image = _TaipyContentImage
    json = "json"
    lov = _TaipyLov
    lov_value = _TaipyLovValue
    number = "number"
    react = "react"
    string = "string"
    string_or_number = "string|number"
    boolean_or_list = "boolean|list"
    number_or_lov_value = "number|lovValue"
    string_list = "stringlist"


def _get_taipy_type(a_type: t.Optional[_AttributeType]) -> t.Optional[t.Type[_TaipyBase]]:
    if isinstance(a_type, _AttributeType) and not isinstance(a_type.value, str):
        return a_type.value
    if a_type == _AttributeType.boolean:
        return _TaipyBool
    elif a_type == _AttributeType.number:
        return _TaipyNumber
    return None



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import typing as t

from flask import g


class _LocalsContext:

    __ctx_g_name = "locals_context"

    def __init__(self) -> None:
        self.__default_module: str = ""
        self._locals_map: t.Dict[str, t.Dict[str, t.Any]] = {}

    def set_default(self, default: t.Dict[str, t.Any]) -> None:
        self.__default_module = default.get("__name__", "")
        self._locals_map[self.__default_module] = default

    def get_default(self) -> t.Dict[str, t.Any]:
        return self._locals_map[self.__default_module]

    def get_all_keys(self) -> t.Set[str]:
        keys = set()
        for v in self._locals_map.values():
            for i in v.keys():
                keys.add(i)
        return keys

    def add(self, context: t.Optional[str], locals_dict: t.Optional[t.Dict[str, t.Any]]):
        if context is not None and locals_dict is not None and context not in self._locals_map:
            self._locals_map[context] = locals_dict

    def set_locals_context(self, context: t.Optional[str]) -> None:
        if context in self._locals_map:
            setattr(g, _LocalsContext.__ctx_g_name, context)

    def get_locals(self) -> t.Dict[str, t.Any]:
        return self.get_default() if (context := self.get_context()) is None else self._locals_map[context]

    def get_context(self) -> t.Optional[str]:
        return getattr(g, _LocalsContext.__ctx_g_name) if hasattr(g, _LocalsContext.__ctx_g_name) else None

    def is_default(self) -> bool:
        return self.get_default() == self.get_locals()

    def reset_locals_context(self) -> None:
        if hasattr(g, _LocalsContext.__ctx_g_name):
            delattr(g, _LocalsContext.__ctx_g_name)



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import typing as t


class _MapDict(object):
    """
    Provide class binding, can utilize getattr, setattr functionality
    Also perform update operation
    """

    __local_vars = ("_dict", "_update_var")

    def __init__(self, dict_import: dict, app_update_var=None):
        self._dict = dict_import
        # Bind app update var function
        self._update_var = app_update_var

    def __len__(self):
        return self._dict.__len__()

    def __length_hint__(self):
        return self._dict.__length_hint__()

    def __getitem__(self, key):
        value = self._dict.__getitem__(key)
        if isinstance(value, dict):
            if self._update_var:
                return _MapDict(value, lambda s, v: self._update_var(f"{key}.{s}", v, from_map_dict=True))
            else:
                return _MapDict(value)
        return value

    def __setitem__(self, key, value):
        if self._update_var:
            self._update_var(key, value)
        else:
            self._dict.__setitem__(key, value)

    def __delitem__(self, key):
        self._dict.__delitem__(key)

    def __missing__(self, key):
        return self._dict.__missing__(key)

    def __iter__(self):
        return self._dict.__iter__()

    def __reversed__(self):
        return self._dict.__reversed__()

    def __contains__(self, item):
        return self._dict.__contains__(item)

    # to be able to use getattr
    def __getattr__(self, attr):
        value = self._dict.__getitem__(attr)
        if isinstance(value, dict):
            if self._update_var:
                return _MapDict(value, lambda s, v: self._update_var(f"{attr}.{s}", v))
            else:
                return _MapDict(value)
        return value

    def __setattr__(self, attr, value):
        if attr in _MapDict.__local_vars:
            super().__setattr__(attr, value)
        else:
            self.__setitem__(attr, value)

    def keys(self):
        return self._dict.keys()

    def values(self):
        return self._dict.values()

    def items(self):
        return self._dict.items()

    def get(self, key: t.Any, default_value: t.Optional[str] = None) -> t.Union[t.Any, None]:
        return self._dict.get(key, default_value)

    def clear(self) -> None:
        self._dict.clear()

    def setdefault(self, key, value=None) -> t.Union[t.Any, None]:
        return self._dict.setdefault(key, value)

    def pop(self, key, default=None) -> t.Any:
        return self._dict.pop(key, default)

    def popitem(self) -> tuple:
        return self._dict.popitem()

    def copy(self) -> _MapDict:
        return _MapDict(self._dict.copy(), self._update_var)

    def update(self, d: dict) -> None:
        current_keys = self.keys()
        for k, v in d.items():
            if k not in current_keys or self[k] != v:
                self.__setitem__(k, v)



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t

import numpy
import pandas as pd

from ..gui import Gui
from .data_format import _DataFormat
from .pandas_data_accessor import _PandasDataAccessor


class _NumpyDataAccessor(_PandasDataAccessor):

    __types = (numpy.ndarray,)

    @staticmethod
    def get_supported_classes() -> t.List[str]:
        return [t.__name__ for t in _NumpyDataAccessor.__types]  # type: ignore

    def __get_dataframe(self, value: t.Any) -> pd.DataFrame:
        return pd.DataFrame(value)

    def get_col_types(self, var_name: str, value: t.Any) -> t.Union[None, t.Dict[str, str]]:  # type: ignore
        if isinstance(value, _NumpyDataAccessor.__types):  # type: ignore
            return super().get_col_types(var_name, self.__get_dataframe(value))
        return None

    def get_data(  # noqa: C901
        self, guiApp: Gui, var_name: str, value: t.Any, payload: t.Dict[str, t.Any], data_format: _DataFormat
    ) -> t.Dict[str, t.Any]:
        if isinstance(value, _NumpyDataAccessor.__types):  # type: ignore
            return super().get_data(guiApp, var_name, self.__get_dataframe(value), payload, data_format)
        return {}



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from markdown.treeprocessors import Treeprocessor

from ..builder import _Builder


class _Postprocessor(Treeprocessor):
    @staticmethod
    def extend(md, gui, priority):
        instance = _Postprocessor(md)
        md.treeprocessors.register(instance, "taipy", priority)
        instance._gui = gui

    def run(self, root):
        MD_PARA_CLASSNAME = "md-para"
        for p in root.iter():
            if p.tag == "p":
                classes = p.get("class")
                classes = f"{MD_PARA_CLASSNAME} {classes}" if classes else MD_PARA_CLASSNAME
                p.set("class", classes)
                p.tag = "div"
            if p != root:
                p.set("key", _Builder._get_key(p.tag))
        return root



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import ast
import inspect
import typing as t
from types import FrameType


def _get_imported_var(frame: FrameType) -> t.List[t.Tuple[str, str, str]]:
    st = ast.parse(inspect.getsource(frame))
    var_list: t.List[t.Tuple[str, str, str]] = []
    for node in ast.walk(st):
        if isinstance(node, ast.ImportFrom):
            # get the imported element as (name, asname, module)
            # ex: from module1 import a as x --> ("a", "x", "module1")
            var_list.extend(
                (
                    child_node.name,
                    child_node.asname if child_node.asname is not None else child_node.name,
                    node.module or "",
                )
                for child_node in node.names
            )
    return var_list



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import ast
import builtins
import re
import typing as t
import warnings

if t.TYPE_CHECKING:
    from ..gui import Gui

from . import (
    _get_client_var_name,
    _get_expr_var_name,
    _getscopeattr,
    _getscopeattr_drill,
    _hasscopeattr,
    _setscopeattr,
    _setscopeattr_drill,
    _TaipyBase,
    _variable_decode,
    _variable_encode,
)


class _Evaluator:

    # Regex to separate content from inside curly braces when evaluating f string expressions
    __EXPR_RE = re.compile(r"\{(([^\}]*)([^\{]*))\}")
    __EXPR_IS_EXPR = re.compile(r"[^\\][{}]")
    __EXPR_IS_EDGE_CASE = re.compile(r"^\s*{([^}]*)}\s*$")
    __EXPR_VALID_VAR_EDGE_CASE = re.compile(r"^([a-zA-Z\.\_0-9]*)$")
    __EXPR_EDGE_CASE_F_STRING = re.compile(r"[\{]*[a-zA-Z_][a-zA-Z0-9_]*:.+")
    __IS_TAIPYEXPR_RE = re.compile(r"TpExPr_(.*)")

    def __init__(self, default_bindings: t.Dict[str, t.Any]) -> None:
        # key = expression, value = hashed value of the expression
        self.__expr_to_hash: t.Dict[str, str] = {}
        # key = hashed value of the expression, value = expression
        self.__hash_to_expr: t.Dict[str, str] = {}
        # key = variable name of the expression, key = list of related expressions
        # ex: {x + y}
        # "x_TPMDL_module1": ["{x + y}"],
        # "y_TPMDL_module1": ["{x + y}"],
        self.__var_to_expr_list: t.Dict[str, t.List[str]] = {}
        # key = expression, value = list of related variables
        # "{x + y}": {"x": "x_TPMDL_module1", "y": "y_TPMDL_module1"}
        self.__expr_to_var_map: t.Dict[str, t.Dict[str, str]] = {}
        # instead of binding everywhere the types
        self.__global_ctx = default_bindings
        # expr to holders
        self.__expr_to_holders: t.Dict[str, t.Set[t.Type[_TaipyBase]]] = {}

    @staticmethod
    def _expr_decode(s: str):
        return str(result[1]) if (result := _Evaluator.__IS_TAIPYEXPR_RE.match(s)) else s

    def get_hash_from_expr(self, expr: str) -> str:
        return self.__expr_to_hash.get(expr, expr)

    def get_expr_from_hash(self, hash: str) -> str:
        return self.__hash_to_expr.get(hash, hash)

    def _is_expression(self, expr: str) -> bool:
        return len(_Evaluator.__EXPR_IS_EXPR.findall(expr)) != 0

    def _fetch_expression_list(self, expr: str) -> t.List:
        return [v[0] for v in _Evaluator.__EXPR_RE.findall(expr)]

    def _analyze_expression(self, gui: Gui, expr: str) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, str]]:
        var_val: t.Dict[str, t.Any] = {}
        var_map: t.Dict[str, str] = {}
        non_vars = list(self.__global_ctx.keys())
        non_vars.extend(dir(builtins))
        # Get a list of expressions (value that has been wrapped in curly braces {}) and find variables to bind
        for e in self._fetch_expression_list(expr):
            var_name = e.split(sep=".")[0]
            st = ast.parse('f"{' + e + '}"' if _Evaluator.__EXPR_EDGE_CASE_F_STRING.match(e) else e)
            args = [arg.arg for node in ast.walk(st) if isinstance(node, ast.arguments) for arg in node.args]
            targets = [
                compr.target.id for node in ast.walk(st) if isinstance(node, ast.ListComp) for compr in node.generators  # type: ignore
            ]
            for node in ast.walk(st):
                if isinstance(node, ast.Name):
                    var_name = node.id.split(sep=".")[0]
                    if var_name not in args and var_name not in targets and var_name not in non_vars:
                        try:
                            encoded_var_name = gui._bind_var(var_name)
                            var_val[var_name] = _getscopeattr_drill(gui, encoded_var_name)
                            var_map[var_name] = encoded_var_name
                        except AttributeError as e:
                            warnings.warn(f"Variable '{var_name}' is not defined (in expression '{expr}'): {e}")
        return var_val, var_map

    def __save_expression(
        self,
        gui: Gui,
        expr: str,
        expr_hash: t.Optional[str],
        expr_evaluated: t.Optional[t.Any],
        var_map: t.Dict[str, str],
    ):
        if expr in self.__expr_to_hash:
            expr_hash = self.__expr_to_hash[expr]
            gui._bind_var_val(expr_hash, expr_evaluated)
            return expr_hash
        if expr_hash is None:
            expr_hash = _get_expr_var_name(expr)
        else:
            # edge case, only a single variable
            expr_hash = f"tpec_{_get_client_var_name(expr)}"
        self.__expr_to_hash[expr] = expr_hash
        gui._bind_var_val(expr_hash, expr_evaluated)
        self.__hash_to_expr[expr_hash] = expr
        for var in var_map.values():
            if var not in self.__global_ctx.keys():
                lst = self.__var_to_expr_list.get(var)
                if lst is None:
                    self.__var_to_expr_list[var] = [expr]
                else:
                    lst.append(expr)
        if expr not in self.__expr_to_var_map:
            self.__expr_to_var_map[expr] = var_map
        return expr_hash

    def evaluate_bind_holder(self, gui: Gui, holder: t.Type[_TaipyBase], expr: str) -> str:
        expr_hash = self.__expr_to_hash.get(expr, "unknownExpr")
        hash_name = self.__get_holder_hash(holder, expr_hash)
        expr_lit = expr.replace("'", "\\'")
        holder_expr = f"{holder.__name__}({expr},'{expr_lit}')"
        self.__evaluate_holder(gui, holder, expr)
        if a_set := self.__expr_to_holders.get(expr):
            a_set.add(holder)
        else:
            self.__expr_to_holders[expr] = {holder}
        self.__expr_to_hash[holder_expr] = hash_name
        # expression is only the first part ...
        expr = expr.split(".")[0]
        self.__expr_to_var_map[holder_expr] = {expr: expr}
        if a_list := self.__var_to_expr_list.get(expr):
            a_list.append(holder_expr)
        else:
            self.__var_to_expr_list[expr] = [holder_expr]
        return hash_name

    def evaluate_holders(self, gui: Gui, expr: str) -> t.List[str]:
        lst = []
        for hld in self.__expr_to_holders.get(expr, []):
            hash = self.__get_holder_hash(hld, self.__expr_to_hash.get(expr, ""))
            self.__evaluate_holder(gui, hld, expr)
            lst.append(hash)
        return lst

    @staticmethod
    def __get_holder_hash(holder: t.Type[_TaipyBase], expr_hash: str) -> str:
        return f"{holder.get_hash()}_{_get_client_var_name(expr_hash)}"

    def __evaluate_holder(self, gui: Gui, holder: t.Type[_TaipyBase], expr: str) -> t.Optional[_TaipyBase]:
        try:
            expr_hash = self.__expr_to_hash.get(expr, "unknownExpr")
            holder_hash = self.__get_holder_hash(holder, expr_hash)
            expr_value = _getscopeattr_drill(gui, expr_hash)
            holder_value = _getscopeattr(gui, holder_hash, None)
            if not isinstance(holder_value, _TaipyBase):
                holder_value = holder(expr_value, expr_hash)
                _setscopeattr(gui, holder_hash, holder_value)
            else:
                holder_value.set(expr_value)
            return holder_value
        except Exception as e:
            warnings.warn(f"Cannot evaluate expression {holder.__name__}({expr_hash},'{expr_hash}') for {expr}: {e}")
        return None

    def evaluate_expr(self, gui: Gui, expr: str) -> t.Any:
        if not self._is_expression(expr):
            return expr
        var_val, var_map = self._analyze_expression(gui, expr)
        expr_hash = None
        is_edge_case = False

        # The expr_string is placed here in case expr get replaced by edge case
        expr_string = 'f"' + expr.replace('"', '\\"') + '"'
        # simplify expression if it only contains var_name
        m = _Evaluator.__EXPR_IS_EDGE_CASE.match(expr)
        if m and not _Evaluator.__EXPR_EDGE_CASE_F_STRING.match(expr):
            expr = m.group(1)
            expr_hash = expr if _Evaluator.__EXPR_VALID_VAR_EDGE_CASE.match(expr) else None
            is_edge_case = True
        # validate whether expression has already been evaluated
        module_name = gui._get_locals_context()
        not_encoded_expr = expr
        expr = f"TpExPr_{_variable_encode(expr, module_name)}"
        if expr in self.__expr_to_hash and _hasscopeattr(gui, self.__expr_to_hash[expr]):
            return self.__expr_to_hash[expr]
        try:
            # evaluate expressions
            ctx: t.Dict[str, t.Any] = {}
            ctx.update(self.__global_ctx)
            # entries in var_val are not always seen (NameError) when passed as locals
            ctx.update(var_val)
            expr_evaluated = eval(not_encoded_expr if is_edge_case else expr_string, ctx)
        except Exception as e:
            warnings.warn(f"Cannot evaluate expression '{not_encoded_expr if is_edge_case else expr_string}': {e}")
            expr_evaluated = None
        # save the expression if it needs to be re-evaluated
        return self.__save_expression(gui, expr, expr_hash, expr_evaluated, var_map)

    def re_evaluate_expr(self, gui: Gui, var_name: str) -> t.Set[str]:
        """
        This function will execute when the _update_var function is handling
        an expression with only a single variable
        """
        modified_vars: t.Set[str] = set()
        # Verify that the current hash is an edge case one (only a single variable inside the original expression)
        if var_name.startswith("tp_"):
            return modified_vars
        expr_original = None
        # if var_name starts with tpec_ --> it is an edge case with modified var
        if var_name.startswith("tpec_"):
            # backup for later reference
            var_name_original = var_name
            expr_original = self.__hash_to_expr[var_name]
            # since this is an edge case --> only 1 item in the dict and that item is the original var
            for v in self.__expr_to_var_map[expr_original].values():
                var_name = v
            # construct correct var_path to reassign values
            var_name_full, _ = _variable_decode(expr_original)
            var_name_full = var_name_full.split(".")
            var_name_full[0] = var_name
            var_name_full = ".".join(var_name_full)
            _setscopeattr_drill(gui, var_name_full, _getscopeattr(gui, var_name_original))
        # otherwise, thar var_name is correct and doesn't require any resolution
        if var_name not in self.__var_to_expr_list:
            # warnings.warn("{var_name} not found")
            return modified_vars
        # refresh expressions and holders
        for expr in self.__var_to_expr_list[var_name]:
            expr_decoded, _ = _variable_decode(expr)
            hash_expr = self.__expr_to_hash.get(expr, "UnknownExpr")
            if expr != var_name and not expr.startswith("_Taipy"):
                expr_var_map = self.__expr_to_var_map.get(expr)  # ["x", "y"]
                if expr_var_map is None:
                    warnings.warn(f"Something is amiss with expression list for {expr}")
                else:
                    eval_dict = {k: _getscopeattr_drill(gui, v) for k, v in expr_var_map.items()}
                    if self._is_expression(expr_decoded):
                        expr_string = 'f"' + _variable_decode(expr)[0].replace('"', '\\"') + '"'
                    else:
                        expr_string = expr_decoded
                    try:
                        ctx: t.Dict[str, t.Any] = {}
                        ctx.update(self.__global_ctx)
                        ctx.update(eval_dict)
                        expr_evaluated = eval(expr_string, ctx)
                        _setscopeattr(gui, hash_expr, expr_evaluated)
                    except Exception as e:
                        warnings.warn(f"Problem evaluating {expr_string}: {e}")
            # refresh holders if any
            for h in self.__expr_to_holders.get(expr, []):
                holder_hash = self.__get_holder_hash(h, self.get_hash_from_expr(expr))
                if holder_hash not in modified_vars:
                    _setscopeattr(gui, holder_hash, self.__evaluate_holder(gui, h, expr))
                    modified_vars.add(holder_hash)
            modified_vars.add(hash_expr)
        return modified_vars



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import warnings
from datetime import date, datetime, time
from json import JSONEncoder
from pathlib import Path

from ..icon import Icon
from ..utils import _date_to_ISO, _MapDict, _TaipyBase


class _TaipyJsonEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, Icon):
            return o._to_dict()
        elif isinstance(o, _MapDict):
            return o._dict
        elif isinstance(o, _TaipyBase):
            return o.get()
        elif isinstance(o, (datetime, date, time)):
            return _date_to_ISO(o)
        elif isinstance(o, Path):
            return str(o)
        try:
            return JSONEncoder.default(self, o)
        except Exception as e:
            warnings.warn(f"JSONEncoder has thrown {e}")
            return None



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import inspect
import typing as t
from abc import ABC, abstractmethod
from os import path
from types import FrameType

from .utils import _filter_locals, _get_module_name_from_frame, _is_in_notebook, _varname_from_content


class Page(ABC):
    """Generic page generator.

    The `Page` class transforms template text into actual pages that can be displayed
    on a Web browser.

    When a page is requested to be displayed, it is converted into HTML
    code that can be sent to the client. All control placeholders are
    replaced by their respective graphical component so you can show
    your application variables and interact with them.
    """

    def __init__(self, content: str, **kwargs) -> None:
        """Initialize a new Page with the indicated content.

        Arguments:
            content (str): The text content or the path to the file holding the text to be transformed.

        If _content_ is a path to a readable file, the file is read entirely as the text template.
        """
        self._content = ""
        self._filepath = ""
        self._frame: t.Optional[FrameType] = None
        if "frame" in kwargs:
            self._frame = kwargs.get("frame")
        else:
            # Get the correct frame from Markdown, Html class by going back 2 stacks
            self._frame = t.cast(FrameType, t.cast(FrameType, inspect.stack()[2].frame))
        self.__process_content(content)

    def __process_content(self, content: str) -> None:
        if path.exists(content) and path.isfile(content):
            with open(t.cast(str, content), "r") as f:
                self._content = f.read()
                # save file path for error handling
                self._filepath = content
        else:
            self._content = content

    def set_content(self, content: str) -> None:
        """Set a new page content.

        Reads the new page content and reinitializes the page to reflect the change.

        !!! important
            This function can only be used an IPython notebook context.

        Arguments:
            content (str): The text content or the path to the file holding the text to be transformed.
                If _content_ is a path to a readable file, the file is read entirely as the text
                template.

        Exceptions:
            RuntimeError: If this method is called outside an IPython notebook context.
        """
        if not _is_in_notebook():
            raise RuntimeError("'set_content()' must be used in an IPython notebook context")
        self.__process_content(content)

    def _get_content_detail(self, gui) -> str:
        if self._filepath:
            return f"in file '{self._filepath}'"
        if varname := _varname_from_content(gui, self._content):
            return f"in variable '{varname}'"
        return ""

    def _get_locals(self) -> t.Optional[t.Dict[str, t.Any]]:
        return None if self._frame is None else _filter_locals(self._frame.f_locals)

    def _get_module_name(self) -> t.Optional[str]:
        return None if self._frame is None else _get_module_name_from_frame(self._frame)

    @abstractmethod
    def render(self, gui) -> str:
        pass



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import typing as t
import warnings
from types import SimpleNamespace


class _DataScopes:

    _GLOBAL_ID = "global"

    def __init__(self) -> None:
        self.__scopes: t.Dict[str, SimpleNamespace] = {}
        self.__scopes[_DataScopes._GLOBAL_ID] = SimpleNamespace()
        self.__single_client = True

    def set_single_client(self, value: bool) -> None:
        self.__single_client = value

    def get_single_client(self) -> bool:
        return self.__single_client

    def get_scope(self, client_id: t.Optional[str]) -> SimpleNamespace:
        if self.__single_client:
            return self.__scopes[_DataScopes._GLOBAL_ID]
        # global context in case request is not registered or client_id is not available (such as in the context of running tests)
        if not client_id:
            warnings.warn("Empty session id, using global scope instead")
            return self.__scopes[_DataScopes._GLOBAL_ID]
        if client_id not in self.__scopes:
            warnings.warn(
                f"session id {client_id} not found in data scope. Taipy will automatically create a scope for this session id but you might have to reload your webpage"
            )
            self.create_scope(client_id)
        return self.__scopes[client_id]

    def get_all_scopes(self) -> t.Dict[str, SimpleNamespace]:
        return self.__scopes

    def broadcast_data(self, name, value):
        if self.__single_client:
            return
        if not self.__client_id:
            for _, v in self.__scopes.items():
                if hasattr(v, name):
                    setattr(v, name, value)

    def create_scope(self, id: str) -> None:
        if self.__single_client:
            return
        if id is None:
            warnings.warn("Empty session id, might be due to unestablished WebSocket connection")
            return
        if id not in self.__scopes:
            self.__scopes[id] = SimpleNamespace()

    def delete_scope(self, id: str) -> None:  # pragma: no cover
        if self.__single_client:
            return
        if id is None:
            warnings.warn("Empty session id, might be due to unestablished WebSocket connection")
            return
        if id in self.__scopes:
            del self.__scopes[id]



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import typing as t
import warnings

from ._page import _Page
from .state import State


class Partial(_Page):
    """Re-usable Page content.

    Partials are used when you need to use a partial page content in different
    and not related pages. This allows not to have to repeat yourself when
    creating your page templates.

    Visual elements such as [`part`](../gui/viselements/part.md),
    [`dialog`](../gui/viselements/dialog.md) or
    [`pane`](../gui/viselements/pane.md) can use Partials.

    Note that `Partial` has no constructor (no `__init__()` method): to create a
    `Partial`, you must call the `Gui.add_partial()^` function.

    Partials can be really handy if you want to modify a section of a page:
    the `update_content()` method dynamically update pages that make use
    of this `Partial` therefore making it easy to change the content of
    any page, at any moment.
    """

    _PARTIALS = "__partials"

    __partials: t.Dict[str, Partial] = {}

    def __init__(self, route: t.Optional[str] = None):
        super().__init__()
        if route is None:
            self._route = f"TaiPy_partials_{len(Partial.__partials)}"
            Partial.__partials[self._route] = self
        else:
            self._route = route

    def update_content(self, state: State, content: str):
        """Update partial content.

        Arguments:
            state (State^): The current user state as received in any callback.
            content (str): The new content to use and display.
        """
        if state and state._gui and callable(state._gui._update_partial):
            state._gui._update_partial(self.__copy(content))
        else:
            warnings.warn("'Partial.update_content()' must be called in the context of a callback")

    def __copy(self, content: str) -> Partial:
        new_partial = Partial(self._route)
        new_partial._renderer = type(self._renderer)(content) if self._renderer is not None else None
        return new_partial



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import inspect
import sys
import typing as t
from types import FrameType, ModuleType


def _get_module_name_from_frame(frame: FrameType):
    return frame.f_globals["__name__"] if "__name__" in frame.f_globals else None


def _get_module_name_from_imported_var(var_name: str, value: t.Any, sub_module_name: str) -> str:
    module_list = sys.modules.copy()
    # the provided sub_module_name are expected to contain only part of the full module_name provided by sys.modules
    # --> we can find potential matched modules based on the sub_module_name
    potential_matched_module = [m for m in module_list.keys() if m.endswith(sub_module_name)]
    for m in potential_matched_module:
        module = module_list[m]
        if hasattr(module, var_name) and getattr(module, var_name) is value:
            return m
    # failed fetching any matched module with variable and value
    return sub_module_name



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import re
import typing as t
import warnings

from ..types import NumberTypes
from ..utils import _get_date_col_str_name, _MapDict, _RE_PD_TYPE


def _add_to_dict_and_get(dico: t.Dict[str, t.Any], key: str, value: t.Any) -> t.Any:
    if key not in dico.keys():
        dico[key] = value
    return dico[key]


def _get_tuple_val(attr: tuple, index: int, default_val: t.Any) -> t.Any:
    return attr[index] if len(attr) > index else default_val


def _get_columns_dict( # noqa: C901
    value: t.Any,
    columns: t.Union[str, t.List[str], t.Tuple[str], t.Dict[str, t.Any], _MapDict],
    col_types: t.Optional[t.Dict[str, str]] = None,
    date_format: t.Optional[str] = None,
    number_format: t.Optional[str] = None,
    opt_columns: t.Optional[t.Set[str]] = None,
):
    if col_types is None:
        return None
    col_types_keys = [str(c) for c in col_types.keys()]
    if isinstance(columns, str):
        columns = [s.strip() for s in columns.split(";")]
    if isinstance(columns, (list, tuple)):
        coldict = {}
        idx = 0
        for col in columns:
            if col not in col_types_keys:
                warnings.warn(
                    f'Error column "{col}" is not present in the dataframe "{value.head(0) if hasattr(value, "head") else value}"'
                )
            else:
                coldict[col] = {"index": idx}
                idx += 1
        if opt_columns is not None:
            for col in opt_columns:
                if col in col_types_keys:
                    coldict[col] = {"index": idx}
                    idx += 1
        columns = coldict
    if isinstance(columns, _MapDict):
        columns = columns._dict
    if not isinstance(columns, dict):
        warnings.warn("Error: columns attributes should be a string, list, tuple or dict")
        columns = {}
    if len(columns) == 0:
        idx = 0
        for col in col_types_keys:
            columns[col] = {"index": idx}
            idx += 1
    idx = 0
    for col, ctype in col_types.items():
        col = str(col)
        if col in columns.keys():
            re_type = _RE_PD_TYPE.match(ctype)
            grps = re_type.groups() if re_type else ()
            ctype = grps[0] if grps else ctype
            columns[col]["type"] = ctype
            columns[col]["dfid"] = col
            if len(grps) > 4 and grps[4]:
                columns[col]["tz"] = grps[4]
            idx = _add_to_dict_and_get(columns[col], "index", idx) + 1
            if ctype == "datetime":
                if date_format:
                    _add_to_dict_and_get(columns[col], "format", date_format)
                columns[_get_date_col_str_name(col_types.keys(), col)] = columns.pop(col)  # type: ignore
            elif number_format and ctype in NumberTypes:
                _add_to_dict_and_get(columns[col], "format", number_format)
    return columns

__RE_INDEXED_DATA = re.compile(r"^(\d+)\/(.*)")


def _get_col_from_indexed(col_name: str, idx: int) -> t.Optional[str]:
    if re_res := __RE_INDEXED_DATA.search(col_name):
        return col_name if str(idx) == re_res.group(1) else None
    return col_name


def _get_idx_from_col(col_name) -> int:
    if re_res := __RE_INDEXED_DATA.search(col_name):
        return int(re_res.group(1))
    return 0


def _to_camel_case(value: str) -> str:
    if not isinstance(value, str):
        raise Exception("_to_camel_case allows only string parameter")

    if len(value) <= 1:
        return value.lower()
    value = value.replace("_", " ").title().replace(" ", "").replace("[", "_").replace("]", "_")
    return value[0].lower() + value[1:]



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t
from pathlib import Path


def _get_non_existent_file_path(dir_path: Path, file_name: str) -> Path:
    if not file_name:
        file_name = "taipy_file.bin"
    file_path = dir_path / file_name
    index = 0
    file_stem = file_path.stem
    file_suffix = file_path.suffix
    while file_path.exists():
        file_path = dir_path / f"{file_stem}.{index}{file_suffix}"
        index += 1
    return file_path



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

# _Page for multipage support
from __future__ import annotations

import logging
import typing as t
import warnings

if t.TYPE_CHECKING:
    from .gui import Gui
    from .renderers import Page


class _Page(object):
    def __init__(self):
        self._rendered_jsx: t.Optional[str] = None
        self._renderer: t.Optional[Page] = None
        self._style: t.Optional[str] = None
        self._route: t.Optional[str] = None
        self._head: t.Optional[list] = None

    def render(self, gui: Gui):
        if self._renderer is None:
            raise RuntimeError(f"Can't render page {self._route}: no renderer found")
        with warnings.catch_warnings(record=True) as w:
            warnings.resetwarnings()
            gui._set_locals_context(self._renderer._get_module_name())
            self._rendered_jsx = self._renderer.render(gui)
            gui._reset_locals_context()
            if w:
                s = "\033[1;31m\n"
                s += (
                    message := f"--- {len(w)} warning(s) were found for page '{'/' if self._route == gui._get_root_page_name() else self._route}' {self._renderer._get_content_detail(gui)} ---\n"
                )
                for i, wm in enumerate(w):
                    s += f" - Warning {i + 1}: {wm.message}\n"
                s += "-" * len(message)
                s += "\033[0m\n"
                logging.warn(s)
        if hasattr(self._renderer, "head"):
            self._head = list(self._renderer.head)  # type: ignore
        # return renderer module_name from frame
        return self._renderer._get_module_name()



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t
from operator import attrgetter

if t.TYPE_CHECKING:
    from ..gui import Gui


def _getscopeattr(gui: "Gui", name: str, *more) -> t.Any:
    if more:
        return getattr(gui._get_data_scope(), name, more[0])
    return getattr(gui._get_data_scope(), name)


def _getscopeattr_drill(gui: "Gui", name: str) -> t.Any:
    return attrgetter(name)(gui._get_data_scope())


def _setscopeattr(gui: "Gui", name: str, value: t.Any):
    setattr(gui._get_data_scope(), name, value)


def _setscopeattr_drill(gui: "Gui", name: str, value: t.Any):
    _attrsetter(gui._get_data_scope(), name, value)


def _hasscopeattr(gui: "Gui", name: str) -> bool:
    return hasattr(gui._get_data_scope(), name)


def _delscopeattr(gui: "Gui", name: str):
    delattr(gui._get_data_scope(), name)


def _attrsetter(obj: object, attr_str: str, value: object) -> None:
    var_name_split = attr_str.split(sep=".")
    for i in range(len(var_name_split) - 1):
        sub_name = var_name_split[i]
        obj = getattr(obj, sub_name)
    setattr(obj, var_name_split[-1], value)



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import argparse
import os
import re
import typing as t
import warnings

import pytz
import tzlocal
from dotenv import dotenv_values

from ._page import _Page
from .partial import Partial

ConfigParameter = t.Literal[
    "port",
    "dark_mode",
    "debug",
    "host",
    "use_reloader",
    "time_zone",
    "propagate",
    "favicon",
    "title",
    "theme",
    "light_theme",
    "dark_theme",
    "use_arrow",
    "system_notification",
    "notification_duration",
    "single_client",
    "ngrok_token",
    "upload_folder",
    "data_url_max_size",
    "flask_log",
    "margin",
    "run_browser",
    "content_security_policy",
    "force_https",
    "watermark",
    "change_delay",
]

Config = t.TypedDict(
    "Config",
    {
        "port": int,
        "dark_mode": bool,
        "debug": bool,
        "host": str,
        "use_reloader": bool,
        "time_zone": t.Union[str, None],
        "propagate": bool,
        "favicon": t.Union[str, None],
        "title": t.Union[str, None],
        "theme": t.Union[t.Dict[str, t.Any], None],
        "light_theme": t.Union[t.Dict[str, t.Any], None],
        "dark_theme": t.Union[t.Dict[str, t.Any], None],
        "use_arrow": bool,
        "system_notification": bool,
        "notification_duration": int,
        "single_client": bool,
        "ngrok_token": str,
        "upload_folder": t.Union[str, None],
        "data_url_max_size": t.Union[int, None],
        "flask_log": bool,
        "margin": t.Union[str, None],
        "run_browser": bool,
        "content_security_policy": t.Optional[dict],
        "force_https": bool,
        "watermark": t.Union[str, None],
        "change_delay": t.Union[int, None],
    },
    total=False,
)


class _Config(object):

    __RE_PORT_NUMBER = re.compile(
        r"^([0-9]{1,4}|[1-5][0-9]{4}|6[0-4][0-9]{3}|65[0-4][0-9]{2}|655[0-2][0-9]|6553[0-5])$"
    )

    def __init__(self):
        self.pages: t.List[_Page] = []
        self.routes: t.List[str] = []
        self.partials: t.List[Partial] = []
        self.partial_routes: t.List[str] = []
        self.config: Config = {}

    def _load(self, config: Config) -> None:
        self.config.update(config)
        # Check that the user timezone configuration setting is valid
        self.get_time_zone()

    def _get_config(self, name: ConfigParameter, default_value: t.Any) -> t.Any:  # pragma: no cover
        if name in self.config and self.config[name] is not None:
            if default_value is not None and not isinstance(self.config[name], type(default_value)):
                try:
                    return type(default_value)(self.config[name])
                except Exception as e:
                    warnings.warn(
                        f'app_config "{name}" value "{self.config[name]}" is not of type {type(default_value)}\n{e}'
                    )
                    return default_value
            return self.config[name]
        return default_value

    def get_time_zone(self) -> t.Optional[str]:
        tz = self.config.get("time_zone")
        if tz is None or tz == "client":
            return tz
        if tz == "server":
            # return python tzlocal IANA Time Zone
            return str(tzlocal.get_localzone())
        # Verify user defined IANA Time Zone is valid
        if tz not in pytz.all_timezones_set:
            raise Exception(
                "Time Zone configuration is not valid. Mistyped 'server', 'client' options or invalid IANA Time Zone"
            )
        # return user defined IANA Time Zone (https://en.wikipedia.org/wiki/List_of_tz_database_time_zones)
        return tz

    def _init_argparse(self):
        parser = argparse.ArgumentParser()
        parser.add_argument("-P", "--port", nargs="?", default="", const="", help="Specify server port")
        parser.add_argument("-H", "--host", nargs="?", default="", const="", help="Specify server host")

        parser.add_argument("--ngrok-token", nargs="?", default="", const="", help="Specify NGROK Authtoken")
        parser.add_argument("--force-https", action="store_true", help="Force HTTPS on all connections")

        debug_group = parser.add_mutually_exclusive_group()
        debug_group.add_argument("--debug", help="Turn on debug", action="store_true")
        debug_group.add_argument("--no-debug", help="Turn off debug", action="store_true")

        reloader_group = parser.add_mutually_exclusive_group()
        reloader_group.add_argument("--use-reloader", help="Auto reload on code changes", action="store_true")
        reloader_group.add_argument("--no-reloader", help="No reload on code changes", action="store_true")

        args, unknown_args = parser.parse_known_args()
        self._handle_argparse(args)

    def _handle_argparse(self, args):  # pragma: no cover
        config = self.config
        if args.port:
            if not _Config.__RE_PORT_NUMBER.match(args.port):
                warnings.warn("Port value for --port option is not valid")
            else:
                config["port"] = int(args.port)
        if args.host:
            config["host"] = args.host
        if args.debug:
            config["debug"] = True
        if args.no_debug:
            config["debug"] = False
        if args.use_reloader:
            config["use_reloader"] = True
        if args.no_reloader:
            config["use_reloader"] = False
        if args.ngrok_token:
            config["ngrok_token"] = args.ngrok_token
        if args.force_https:
            config["force_https"] = True

    def _build_config(self, root_dir, env_filename, kwargs):  # pragma: no cover
        config = self.config
        env_file_abs_path = env_filename if os.path.isabs(env_filename) else os.path.join(root_dir, env_filename)
        # Load keyword arguments
        for key, value in kwargs.items():
            key = key.lower()
            if value is not None and key in config:
                try:
                    config[key] = value if config[key] is None else type(config[key])(value)  # type: ignore
                except Exception as e:
                    warnings.warn(
                        f"Invalid keyword arguments value in Gui.run {key} - {value}. Unable to parse value to the correct type.\n{e}"
                    )
        # Load config from env file
        if os.path.isfile(env_file_abs_path):
            for key, value in dotenv_values(env_file_abs_path).items():
                key = key.lower()
                if value is not None and key in config:
                    try:
                        config[key] = value if config[key] is None else type(config[key])(value)  # type: ignore
                    except Exception as e:
                        warnings.warn(
                            f"Invalid env value in Gui.run: {key} - {value}. Unable to parse value to the correct type.\n{e}"
                        )
        # Load from system arguments
        self._init_argparse()



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import re

from markdown.blockprocessors import BlockProcessor

from .factory import _MarkdownFactory


class _StartBlockProcessor(BlockProcessor):
    __RE_FENCE_START = re.compile(
        _MarkdownFactory._TAIPY_START + r"([a-zA-Z][\.a-zA-Z_$0-9]*)\.start(.*?)" + _MarkdownFactory._TAIPY_END
    )  # start line
    __RE_OTHER_FENCE = re.compile(
        _MarkdownFactory._TAIPY_START + r"([a-zA-Z][\.a-zA-Z_$0-9]*)\.(start|end)(.*?)" + _MarkdownFactory._TAIPY_END
    )  # start or end tag

    @staticmethod
    def extend(md, gui, priority):
        instance = _StartBlockProcessor(md.parser)
        md.parser.blockprocessors.register(instance, "taipy", priority)
        instance._gui = gui

    def test(self, parent, block):
        return re.match(_StartBlockProcessor.__RE_FENCE_START, block)

    def run(self, parent, blocks):
        original_block = blocks[0]
        original_match = re.search(_StartBlockProcessor.__RE_FENCE_START, original_block)
        blocks[0] = re.sub(_StartBlockProcessor.__RE_FENCE_START, "", blocks[0], 1)
        tag = original_match.group(1)
        queue = [tag]
        # Find block with ending fence
        for block_num, block in enumerate(blocks):
            matches = re.findall(_StartBlockProcessor.__RE_OTHER_FENCE, block)
            for match in matches:
                if queue[-1] == match[0] and match[1] == "end":
                    queue.pop()
                elif match[1] == "start":
                    queue.append(match[0])
            if not queue:
                # remove end fence
                blocks[block_num] = re.sub(
                    _MarkdownFactory._TAIPY_START + tag + r"\.end(.*?)" + _MarkdownFactory._TAIPY_END,
                    "",
                    block,
                    1,
                )
                # render fenced area inside a new div
                e = _MarkdownFactory.create_element(self._gui, original_match.group(1), original_match.group(2))
                parent.append(e)
                # parse inside blocks
                self.parser.parseBlocks(e, blocks[:block_num + 1])
                # remove used blocks
                del blocks[:block_num + 1]
                return True  # or could have had no return statement
        # No closing marker!  Restore and do nothing
        blocks[0] = original_block
        return False  # equivalent to our test() routine returning False



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import re
import typing as t
import warnings
from html.parser import HTMLParser

from .factory import _HtmlFactory


class _TaipyHTMLParser(HTMLParser):

    __TAIPY_NAMESPACE_RE = re.compile(r"taipy:([a-zA-Z\_]*)")

    def __init__(self, gui):
        super().__init__()
        self._gui = gui
        self.body = ""
        self.head = []
        self.taipy_tag = None
        self.tag_mapping = {}
        self.is_body = True
        self.head_tag = None
        self._line_count = 0
        self._tag_queue = []

    # @override
    def handle_starttag(self, tag, props) -> None:
        self._tag_queue.append((tag, self._line_count))
        if tag == "html":
            return
        if self.head_tag is not None:
            self.head.append(self.head_tag)
            self.head_tag = None
        if self.taipy_tag is not None:
            self.parse_taipy_tag()
        if tag == "head":
            self.is_body = False
        elif tag == "body":
            self.is_body = True
        elif m := self.__TAIPY_NAMESPACE_RE.match(tag):
            self.taipy_tag = _TaipyTag(m.group(1), props)
        elif not self.is_body:
            head_props = {prop[0]: prop[1] for prop in props}
            self.head_tag = {"tag": tag, "props": head_props, "content": ""}
        else:
            self.append_data(str(self.get_starttag_text()))

    # @override
    def handle_data(self, data: str) -> None:
        data = data.strip()
        if data and self.taipy_tag is not None and self.taipy_tag.set_value(data):
            self.parse_taipy_tag()
        elif not self.is_body and self.head_tag is not None:
            self.head_tag["content"] = data
        else:
            self.append_data(data)

    # @override
    def handle_endtag(self, tag) -> None:
        if not self._tag_queue:
            warnings.warn(f"Closing '{tag}' at line {self._line_count} is missing an opening tag")
        else:
            opening_tag, opening_tag_line = self._tag_queue.pop()
            if opening_tag != tag:
                warnings.warn(
                    f"Opening tag '{opening_tag}' at line {opening_tag_line} has no matching closing tag '{tag}' at line {self._line_count}"
                )
        if tag in ["head", "body", "html"]:
            return
        if self.taipy_tag is not None:
            self.parse_taipy_tag()
        if not self.is_body:
            self.head.append(self.head_tag)
            self.head_tag = None
        elif tag in self.tag_mapping:
            self.append_data(f"</{self.tag_mapping[tag]}>")
        else:
            self.append_data(f"</{tag}>")

    def append_data(self, data: str) -> None:
        if self.is_body:
            self.body += data

    def parse_taipy_tag(self) -> None:
        tp_string, tp_element_name = self.taipy_tag.parse(self._gui)
        self.append_data(tp_string)
        self.tag_mapping[f"taipy:{self.taipy_tag.control_type}"] = tp_element_name
        self.taipy_tag = None

    def get_jsx(self) -> str:
        return self.body

    def feed_data(self, data: str):
        data_lines = data.split("\n")
        for line, data_line in enumerate(data_lines):
            self._line_count = line + 1
            self.feed(data_line)
        while self._tag_queue:
            opening_tag, opening_tag_line = self._tag_queue.pop()
            warnings.warn(f"Opening tag '{opening_tag}' at line {opening_tag_line} has no matching closing tag")


class _TaipyTag(object):
    def __init__(self, tag_name: str, properties: t.List[t.Tuple[str, str]]) -> None:
        self.control_type = tag_name
        self.properties = {prop[0]: prop[1] for prop in properties}
        self.has_set_value = False

    def set_value(self, value: str) -> bool:
        if self.has_set_value:
            return False
        property_name = _HtmlFactory.get_default_property_name(self.control_type)
        # Set property only if it is not already defined
        if property_name and property_name not in self.properties.keys():
            self.properties[property_name] = value
        self.has_set_value = True
        return True

    def parse(self, gui) -> t.Tuple[str, str]:
        for k, v in self.properties.items():
            self.properties[k] = v if v is not None else "true"
        # allow usage of 'class' property in html taipy tag
        if "class" in self.properties and "class_name" not in self.properties:
            self.properties["class_name"] = self.properties["class"]
        return _HtmlFactory.create_element(gui, self.control_type, self.properties)



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t
from operator import attrgetter

from .utils._attributes import _attrsetter

if t.TYPE_CHECKING:
    from .gui import Gui


class State:
    """Accessor to the bound variables from callbacks.

    `State` is used when you need to access the value of variables
    bound to visual elements (see [Binding](../gui/binding.md)).<br/>
    Because each browser connected to the application server may represent and
    modify any variable at any moment as the result of user interaction, each
    connection holds its own set of variables along with their values. We call
    the set of these the application variables the application _state_, as seen
    by a given client.

    Each callback (see [Callbacks](../gui/callbacks.md))) receives a specific
    instance of the `State` class, where you can find all the variables bound to
    visual elements in your application.

    Note that `State` also is a Python Context Manager: In situations where you
    have several variables to update, it is more clear and more efficient to assign
    the variable values in a `with` construct:
    ```py
    def my_callback(state, ...):
      ...
      with state as s:
        s.var1 = value1
        s.var2 = value2
      ...
    ```

    You cannot set a variable in the context of a lambda function because Python
    prevents any use of the assignment operator.<br/>
    You can, however, use the `assign()` method on the state that the lambda function
    receives, so you can work around this limitation:

    Here is how you could define a button that changes the value of a variable
    directly in the Markdown code:
    ```
       <|Set variable|button|on_action={lambda s: s.assign("var_name", new_value}|>
    ```
    This would be strictly similar to the Markdown line:
    ```
       <|Set variable|button|on_action=change_variable|>
    ```
    with the Python code:
    ```py
    def change_variable(state):
        state.var_name = new_value
    ```
    """

    __gui_attr = "_gui"
    __attrs = (__gui_attr, "_user_var_list")
    __methods = ("assign", "_get_placeholder", "_set_placeholder", "_get_gui_attr", "_get_placeholder_attrs")
    __placeholder_attrs = ("_taipy_p1",)
    __excluded_attrs = __attrs + __methods + __placeholder_attrs

    def __init__(self, gui: "Gui", var_list: t.Iterable[str]) -> None:
        super().__setattr__(State.__attrs[1], list(State.__filter_var_list(var_list, State.__excluded_attrs)))
        super().__setattr__(State.__attrs[0], gui)

    @staticmethod
    def __filter_var_list(var_list: t.Iterable[str], excluded_attrs: t.Iterable[str]) -> t.Iterable[str]:
        return filter(lambda n: n not in excluded_attrs, var_list)

    def __getattribute__(self, name: str) -> t.Any:
        if name in State.__methods:
            return super().__getattribute__(name)
        gui = super().__getattribute__(State.__gui_attr)
        if name == State.__gui_attr:
            return gui
        if name in State.__excluded_attrs:
            raise AttributeError(f"Variable '{name}' is protected and is not accessible.")
        if name not in super().__getattribute__(State.__attrs[1]):
            raise AttributeError(f"Variable '{name}' is not defined.")
        encoded_name = gui._bind_var(name)
        return getattr(gui._bindings(), encoded_name)

    def __setattr__(self, name: str, value: t.Any) -> None:
        if name not in super().__getattribute__(State.__attrs[1]):
            raise AttributeError(f"Variable '{name}' is not accessible.")
        gui = super().__getattribute__(State.__gui_attr)
        encoded_name = gui._bind_var(name)
        setattr(gui._bindings(), encoded_name, value)

    def _get_placeholder(self, name: str):
        if name in State.__placeholder_attrs:
            try:
                return super().__getattribute__(name)
            except AttributeError:
                return None
        return None

    def _set_placeholder(self, name: str, value: t.Any):
        if name in State.__placeholder_attrs:
            super().__setattr__(name, value)

    def _get_gui_attr(self):
        return State.__gui_attr

    def _get_placeholder_attrs(self):
        return State.__placeholder_attrs

    def assign(self, name: str, value: t.Any) -> t.Any:
        """Assign a value to a state variable.

        This should be used only from within a lambda function used
        as a callback in a visual element.

        Args:
            name (str): The variable name to assign to.
            value (Any): The new variable value.

        Returns:
            Any: The previous value of the variable.
        """
        val = attrgetter(name)(self)
        _attrsetter(self, name, value)
        return val

    def __enter__(self):
        super().__getattribute__(State.__attrs[0]).__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return super().__getattribute__(State.__attrs[0]).__exit__(exc_type, exc_value, traceback)



#!/usr/bin/env python

# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""The setup script."""

import os
from pathlib import Path

from setuptools import find_packages, setup
from setuptools.command.build_py import build_py

with open("README.md") as readme_file:
    readme = readme_file.read()


requirements = [
    "flask>=2.1,<3.0",
    "flask-cors>=3.0.10,<4.0",
    "flask-socketio>=5.1.1,<6.0",
    "markdown>=3.3.4,<4.0",
    "pandas>=1.3.4,<2.0",
    "python-dotenv>=0.19,<0.21",
    "pytz>=2021.3,<2022.2",
    "simple-websocket>=0.3,<0.6",
    "tzlocal>=3.0,<5.0",
    "backports.zoneinfo>=0.2.1,<0.3;python_version<'3.9'",
    "flask-talisman>=1.0,<2.0",
]

test_requirements = ["pytest>=3.8"]

extras_require = {
    "ngrok": ["pyngrok>=5.1,<6.0"],
    "image": ["python-magic>=0.4.24,<0.5;platform_system!='Windows'", "python-magic-bin>=0.4.14,<0.5;platform_system=='Windows'"],
    "rdp": ["rdp>=0.8"],
    "arrow": ["pyarrow>=7.0,<9.0"],
}


def _build_webapp():
    already_exists = Path(f"./taipy/gui/webapp/index.html").exists()
    if not already_exists:
        os.system("cd gui && npm ci && npm run build")


class NPMInstall(build_py):
    def run(self):
        _build_webapp()
        build_py.run(self)


setup(
    author="Avaiga",
    author_email="dev@taipy.io",
    python_requires=">=3.8",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="A 360° open-source platform from Python pilots to production-ready web apps.",
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    license="Apache License 2.0",
    include_package_data=True,
    keywords="taipy-gui",
    name="taipy-gui",
    packages=find_packages(include=["taipy", "taipy.gui", "taipy.gui.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/avaiga/taipy-gui",
    version="1.1.0",
    zip_safe=False,
    extras_require=extras_require,
    cmdclass={"build_py": NPMInstall},
)



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t

from ..page import Page
from ._html import _TaipyHTMLParser

if t.TYPE_CHECKING:
    from ..gui import Gui


class _EmptyPage(Page):
    def __init__(self) -> None:
        super().__init__("<PageContent />")

    def render(self, gui) -> str:
        return self._content


class Markdown(Page):
    """
    Page generator for _Markdown_ text.

    Taipy can use Markdown text to create pages that are the base of
    user interfaces.

    You can find details on the Taipy Markdown-specific syntax and how to add
    Taipy Visual Elements in the [section on HTML](../../gui/pages/#using-markdown)
    of the User Manual.
    """

    def __init__(self, content: str, **kwargs) -> None:
        """Initialize a new `Markdown` page.

        Arguments:
            content (str): The text content or the path to the file holding the Markdown text
                to be transformed.<br/>
                If _content_ is a path to a readable file, the file is read as the Markdown
                template content.
        """
        super().__init__(content, **kwargs)

    # Generate JSX from Markdown
    def render(self, gui) -> str:
        return gui._markdown.convert(self._content)


class Html(Page):
    """
    Page generator for _HTML_ text.

    Taipy can use HTML code to create pages that are the base of
    user interfaces.

    You can find details on HTML-specific constructs and how to add
    Taipy Visual Elements in the [section on HTML](../../gui/pages/#using-html)
    of the User Manual.
    """

    def __init__(self, content: str, **kwargs) -> None:
        """Initialize a new `Html` page.

        Arguments:
            content (str): The text content or the path to the file holding the HTML text to
                be transformed.<br/>
                If _content_ is a path to a readable file, the file is read as the HTML
                template content.
        """
        super().__init__(content, **kwargs)
        self.head = None

    # Modify path routes
    def modify_taipy_base_url(self, base_url):
        self._content = self._content.replace("{{taipy_base_url}}", f"/{base_url}")

    # Generate JSX from HTML
    def render(self, gui) -> str:
        parser = _TaipyHTMLParser(gui)
        parser.feed_data(self._content)
        self.head = parser.head
        return parser.get_jsx()



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import re
import typing as t

__expr_var_name_index: t.Dict[str, int] = {}
_RE_NOT_IN_VAR_NAME = r"[^A-Za-z0-9]+"


def _get_expr_var_name(expr: str) -> str:
    var_name = re.sub(_RE_NOT_IN_VAR_NAME, "_", expr)
    index = 0
    if var_name in __expr_var_name_index.keys():
        index = __expr_var_name_index[var_name]
    __expr_var_name_index[var_name] = index + 1
    return f"tp_{var_name}_{index}"


def _reset_expr_var_name():
    __expr_var_name_index.clear()



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t
import warnings
from importlib import util

import numpy as np
import pandas as pd

from ..gui import Gui
from ..utils import _get_date_col_str_name, _RE_PD_TYPE
from .data_accessor import _DataAccessor
from .data_format import _DataFormat

_has_arrow_module = False
if util.find_spec("pyarrow"):
    _has_arrow_module = True
    import pyarrow as pa


class _PandasDataAccessor(_DataAccessor):

    __types = (pd.DataFrame,)

    @staticmethod
    def get_supported_classes() -> t.List[str]:
        return [t.__name__ for t in _PandasDataAccessor.__types]

    @staticmethod
    def __style_function(
        row: pd.Series, gui: Gui, column_name: t.Optional[str], user_function: t.Callable, function_name: str
    ) -> str:  # pragma: no cover
        args = []
        if column_name:
            args.append(row[column_name])
        args.extend((row.name, row))
        if column_name:
            args.append(column_name)
        try:
            return str(gui._call_function_with_state(user_function, args))
        except Exception as e:
            warnings.warn(f"Exception raised when calling user style function {function_name}\n{e}")
        return ""

    def __build_transferred_cols(
        self, gui: Gui, payload_cols: t.Any, data: pd.DataFrame, styles: t.Optional[t.Dict[str, str]] = None
    ) -> pd.DataFrame:
        if isinstance(payload_cols, list) and len(payload_cols):
            col_types = data.dtypes[data.dtypes.index.astype(str).isin(payload_cols)]
        else:
            col_types = data.dtypes
        cols = col_types.index.astype(str).tolist()
        is_copied = False
        if styles:
            # copy the df so that we don't "mess" with the user's data
            data = data.copy()
            is_copied = True
            for k, v in styles.items():
                applied = False
                func = gui._get_user_function(v)
                if callable(func):
                    applied = self.__apply_user_function(gui, func, k if k in cols else None, v, data)
                if not applied:
                    data[v] = v
                cols.append(v)
        # deal with dates
        datecols = col_types[col_types.astype(str).str.startswith("datetime")].index.tolist()
        if len(datecols) != 0:
            if not is_copied:
                # copy the df so that we don't "mess" with the user's data
                data = data.copy()
            tz = Gui._get_timezone()
            for col in datecols:
                newcol = _get_date_col_str_name(cols, col)
                cols.append(newcol)
                re_type = _RE_PD_TYPE.match(str(col_types[col]))
                grps = re_type.groups() if re_type else ()
                if len(grps) > 4 and grps[4]:
                    data[newcol] = data[col].dt.tz_convert("UTC").dt.strftime(_DataAccessor._WS_DATE_FORMAT).astype(str)
                else:
                    data[newcol] = data[col].dt.tz_localize(tz).dt.tz_convert("UTC").dt.strftime(_DataAccessor._WS_DATE_FORMAT).astype(str)
            # remove the date columns from the list of columns
            cols = list(set(cols) - set(datecols))
        data = data.loc[:, data.dtypes[data.dtypes.index.astype(str).isin(cols)].index]
        return data

    def __apply_user_function(
        self, gui: Gui, user_function: t.Callable, column_name: t.Optional[str], function_name: str, data: pd.DataFrame
    ):
        arg_count = user_function.__code__.co_argcount
        if arg_count > 1:
            data[function_name] = data.apply(
                _PandasDataAccessor.__style_function,
                axis=1,
                args=(gui, column_name, user_function, function_name),
            )
            return True
        else:
            warnings.warn(f"Style function {function_name} should take at least a value as second argument")
        return False

    def __format_data(
        self,
        data: pd.DataFrame,
        data_format: _DataFormat,
        orient: str,
        start: t.Optional[int] = None,
        rowcount: t.Optional[int] = None,
        data_extraction: t.Optional[bool] = None,
        handle_nan: t.Optional[bool] = False,
    ) -> t.Dict[str, t.Any]:
        ret: t.Dict[str, t.Any] = {
            "format": str(data_format.value),
        }
        if rowcount is not None:
            ret["rowcount"] = rowcount
        if start is not None:
            ret["start"] = start
        if data_extraction is not None:
            ret["dataExtraction"] = data_extraction  # Extract data out of dictionary on frontend
        if data_format == _DataFormat.APACHE_ARROW and _has_arrow_module:
            # Convert from pandas to Arrow
            table = pa.Table.from_pandas(data)
            # Create sink buffer stream
            sink = pa.BufferOutputStream()
            # Create Stream writer
            writer = pa.ipc.new_stream(sink, table.schema)
            # Write data to table
            writer.write_table(table)
            writer.close()
            # end buffer stream
            buf = sink.getvalue()
            # convert buffer to python bytes and return
            ret["data"] = buf.to_pybytes()
            ret["orient"] = orient
        elif data_format == _DataFormat.APACHE_ARROW:
            raise RuntimeError("Cannot use Arrow as pyarrow package is not installed")
        else:
            # workaround for python built in json encoder that does not yet support ignore_nan
            ret["data"] = data.replace([np.nan], ["NaN" if handle_nan else None]).to_dict(orient=orient)  # type: ignore
        return ret

    def get_col_types(self, var_name: str, value: t.Any) -> t.Union[None, t.Dict[str, str]]:  # type: ignore
        if isinstance(value, _PandasDataAccessor.__types):
            return value.dtypes.apply(lambda x: x.name).to_dict()
        elif isinstance(value, list):
            ret_dict: t.Dict[str, str] = {}
            for i, v in enumerate(value):
                ret_dict.update({f"{i}/{k}": v for k, v in v.dtypes.apply(lambda x: x.name).items()})
            return ret_dict
        return None

    def __get_data(  # noqa: C901
        self,
        gui: Gui,
        var_name: str,
        value: pd.DataFrame,
        payload: t.Dict[str, t.Any],
        data_format: _DataFormat,
        col_prefix: t.Optional[str] = "",
    ) -> t.Dict[str, t.Any]:
        columns = payload.get("columns", [])
        if col_prefix:
            columns = [c[len(col_prefix):] if c.startswith(col_prefix) else c for c in columns]
        ret_payload = {"pagekey": payload.get("pagekey", "unknown page")}
        paged = not payload.get("alldata", False)
        if paged:
            aggregates = payload.get("aggregates")
            applies = payload.get("applies")
            if isinstance(aggregates, list) and len(aggregates) and isinstance(applies, dict):
                applies_with_fn = {
                    k: v if v in gui._aggregate_functions else gui._get_user_function(v) for k, v in applies.items()
                }

                for col in columns:
                    if col not in applies_with_fn.keys():
                        applies_with_fn[col] = "first"
                try:
                    value = value.groupby(aggregates).agg(applies_with_fn)
                except Exception:
                    warnings.warn(f"Cannot aggregate {var_name} with groupby {aggregates} and aggregates {applies}")
            inf = payload.get("infinite")
            if inf is not None:
                ret_payload["infinite"] = inf
            # real number of rows is needed to calculate the number of pages
            rowcount = len(value)
            # here we'll deal with start and end values from payload if present
            if isinstance(payload["start"], int):
                start = int(payload["start"])
            else:
                try:
                    start = int(str(payload["start"]), base=10)
                except Exception:
                    warnings.warn(f'start should be an int value {payload["start"]}')
                    start = 0
            if isinstance(payload["end"], int):
                end = int(payload["end"])
            else:
                try:
                    end = int(str(payload["end"]), base=10)
                except Exception:
                    end = -1
            if start < 0 or start >= rowcount:
                start = 0
            if end < 0 or end >= rowcount:
                end = rowcount - 1
            # deal with sort
            order_by = payload.get("orderby")
            if isinstance(order_by, str) and len(order_by):
                new_indexes = value[order_by].values.argsort(axis=0)
                if payload.get("sort") == "desc":
                    # reverse order
                    new_indexes = new_indexes[::-1]
                new_indexes = new_indexes[slice(start, end + 1)]
            else:
                new_indexes = slice(start, end + 1)  # type: ignore
            value = self.__build_transferred_cols(gui, columns, value.iloc[new_indexes], styles=payload.get("styles"))
            dictret = self.__format_data(
                value, data_format, "records", start, rowcount, handle_nan=payload.get("handlenan", False)
            )
        else:
            ret_payload["alldata"] = True
            nb_rows_max = payload.get("width")
            # view with the requested columns
            if nb_rows_max and nb_rows_max < len(value) / 2:
                value = value.copy()
                value["tp_index"] = value.index
                # we need to be more clever than this :-)
                value = value.iloc[:: (len(value) // nb_rows_max)]
            value = self.__build_transferred_cols(gui, columns, value)
            dictret = self.__format_data(value, data_format, "list", data_extraction=True)
        ret_payload["value"] = dictret
        return ret_payload

    def get_data(
        self, gui: Gui, var_name: str, value: t.Any, payload: t.Dict[str, t.Any], data_format: _DataFormat
    ) -> t.Dict[str, t.Any]:
        if isinstance(value, list):
            is_chart = payload.get("alldata", False)
            if is_chart:
                ret_payload = {
                    "alldata": True,
                    "value": {"multi": True},
                    "pagekey": payload.get("pagekey", "unknown page"),
                }
                data = []
                for i, v in enumerate(value):
                    ret = (
                        self.__get_data(gui, var_name, v, payload, data_format, f"{i}/")
                        if isinstance(v, pd.DataFrame)
                        else {}
                    )
                    ret_val = ret.get("value", {})
                    data.append(ret_val.pop("data", None))
                    ret_payload.get("value", {}).update(ret_val)
                ret_payload["value"]["data"] = data
                return ret_payload
            else:
                value = value[0]
        if isinstance(value, _PandasDataAccessor.__types):
            return self.__get_data(gui, var_name, value, payload, data_format)
        return {}



# Copyright 2022 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""# Taipy Graphical User Interface generator.

The Taipy Gui package provides User Interface generation based on
page templates. It can run a Web server that a Web browser can connect
to.

The pages are generated by a Web server that allows Web clients to
connect, display and interact with the page content through visual
elements.

Each page can contain regular text and images, as well
as Taipy controls that are typically linked to some
value that is managed by the whole Taipy application.

Here is how you can create your first Taipy User Interface:

   - Create a Python source file.
     Copy these two lines into a file called _taipy_app.py_.
     ```py
     from taipy import Gui
     Gui("# Hello Taipy!").run()
     ```
   - Install Taipy:
     ```
     pip install taipy
     ```
   - Run your application:
     ```
     python taipy_app.py
     ```

Your browser opens a new page, showing the content of your graphical
application.

!!! Note "Optional packages"

    There are Python packages that you can install in your environment to
    add functionality to Taipy GUI:

    - [`python-magic`](https://pypi.org/project/python-magic/): identifies image format from
      byte buffers so the [`image`](../../gui/viselements/image.md) control can display them,
      and so that [`file_download`](../../gui/viselements/file_download.md) can request
      the browser to display the image content when relevant.<br/>
      You can install that package with the regular `pip install python-magic` command
      (then potentially `pip install python-magic` on Windows),
      or install Taipy GUI using: `pip install taipy-gui[image]`.
    - [`pyarrow`](https://pypi.org/project/pyarrow/): can improve the performance of your
      application by reducing the volume of data transferred between the Web server and the
      clients. This is relevant if your application uses large tabular data.<br/>
      You can install that package with the regular `pip install pyarrow` command,
      or install Taipy GUI using: `pip install taipy-gui[arrow]`.

"""

from .gui import Gui
from .gui_actions import download, hold_control, navigate, notify, resume_control, get_context_id, invoke_state_callback
from .icon import Icon
from .renderers import Html, Markdown
from .state import State
